# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Picklable read-only I/O classes --- :mod:`MDAnalysis.lib.picklable_file_io`
===========================================================================

Provide with an interface for pickling read-only IO file object.
These classes are used for further pickling :class:`MDAnalysis.core.universe`
in a object composition approach.

.. autoclass:: FileIOPicklable
   :members:

.. autoclass:: BufferIOPicklable
   :members:

.. autoclass:: TextIOPicklable
   :members:

.. autoclass:: BZ2Picklable
   :members:

.. autoclass:: GzipPicklable
   :members:

.. autofunction:: pickle_open

.. autofunction:: bz2_pickle_open

.. autofunction:: gzip_pickle_open


.. versionadded:: 2.0.0
"""
import io
import os

import bz2
import gzip


class FileIOPicklable(io.FileIO):
    """File object (read-only) that can be pickled.

    This class provides a file-like object (as returned by :func:`open`,
    namely :class:`io.FileIO`) that, unlike standard Python file objects,
    can be pickled. Only read mode is supported.

    When the file is pickled, filename and position of the open file handle in
    the file are saved. On unpickling, the file is opened by filename,
    and the file is seeked to the saved position.
    This means that for a successful unpickle, the original file still has to
    be accessible with its filename.

    Note
    ----
    This class only supports reading files in binary mode. If you need to open
    a file in text mode, use the :func:`pickle_open`.

    Parameters
    ----------
    name : str
        either a text or byte string giving the name (and the path
        if the file isn't in the current working directory) of the file to
        be opened.
    mode : str
        only reading ('r') mode works. It exists to be consistent
        with a wider API.

    Example
    -------
    ::

        >>> import pickle
        >>> from MDAnalysis.tests.datafiles import PDB
        >>> file = FileIOPicklable(PDB)
        >>> _ = file.readline()
        >>> file_pickled = pickle.loads(pickle.dumps(file))
        >>> print(file.tell(), file_pickled.tell())
        55 55

    See Also
    ---------
    TextIOPicklable
    BufferIOPicklable


    .. versionadded:: 2.0.0
    """

    def __init__(self, name, mode="r"):
        self._mode = mode
        super().__init__(name, mode)

    def __setstate__(self, state):
        name = state["name_val"]
        self.__init__(name, mode="r")
        try:
            self.seek(state["tell_val"])
        except KeyError:
            pass

    def __reduce_ex__(self, prot):
        if self._mode != "r":
            raise RuntimeError(
                "Can only pickle files that were opened "
                "in read mode, not {}".format(self._mode)
            )
        return (
            self.__class__,
            (self.name, self._mode),
            {"name_val": self.name, "tell_val": self.tell()},
        )


class BufferIOPicklable(io.BufferedReader):
    """A picklable buffer object for read-only FileIO object.

    This class provides a buffered :class:`io.BufferedReader`
    that can be pickled.
    Note that this only works in read mode.

    Parameters
    ----------
    raw : FileIO object

    Example
    -------
    ::

        file = FileIOPicklable('filename')
        buffer_wrapped = BufferIOPicklable(file)

    See Also
    ---------
    FileIOPicklable
    TextIOPicklable


    .. versionadded:: 2.0.0
    """

    def __init__(self, raw):
        super().__init__(raw)
        self.raw_class = raw.__class__

    def __setstate__(self, state):
        raw_class = state["raw_class"]
        name = state["name_val"]
        raw = raw_class(name)
        self.__init__(raw)
        self.seek(state["tell_val"])

    def __reduce_ex__(self, prot):
        # don't ask, for Python 3.12+ see:
        # https://github.com/python/cpython/pull/104370
        return (
            self.raw_class,
            (self.name,),
            {
                "raw_class": self.raw_class,
                "name_val": self.name,
                "tell_val": self.tell(),
            },
        )


class TextIOPicklable(io.TextIOWrapper):
    """Character and line based picklable file-like object.

    This class provides a file-like :class:`io.TextIOWrapper` object that can
    be pickled. Note that this only works in read mode.

    Parameters
    ----------
    raw : FileIO object

    Example
    -------
    ::

        file = FileIOPicklable('filename')
        text_wrapped = TextIOPicklable(file)

    See Also
    ---------
    FileIOPicklable
    BufferIOPicklable


    .. versionadded:: 2.0.0
    .. versionchanged:: 2.8.0
       The raw class instance instead of the class name
       that is wrapped inside will be serialized.
       After deserialization, the current position is no longer reset
       so `universe.trajectory[i]` is not needed to seek to the
       original position.
    """

    def __init__(self, raw):
        super().__init__(raw)
        self.raw_class = raw.__class__

    def __setstate__(self, args):
        raw_class = args["raw_class"]
        name = args["name_val"]
        tell = args["tell_val"]
        # raw_class is used for further expansion this functionality to
        # Gzip files, which also requires a text wrapper.
        raw = raw_class(name)
        self.__init__(raw)
        if tell is not None:
            self.seek(tell)

    def __reduce_ex__(self, prot):
        try:
            curr_loc = self.tell()
        # some readers (e.g. GMS) disable tell() due to using next()
        except OSError:
            curr_loc = None
        try:
            name = self.name
        except AttributeError:
            # This is kind of ugly--BZ2File does not save its name.
            name = self.buffer._fp.name
        return (
            self.__class__.__new__,
            (self.__class__,),
            {
                "raw_class": self.raw_class,
                "name_val": name,
                "tell_val": curr_loc,
            },
        )


class BZ2Picklable(bz2.BZ2File):
    """File object (read-only) for bzip2 (de)compression that can be pickled.

    This class provides a file-like object (as returned by :func:`bz2.open`,
    namely :class:`bz2.BZ2File`) that, unlike standard Python file objects,
    can be pickled. Only read mode is supported.

    When the file is pickled, filename and position of the open file handle in
    the file are saved. On unpickling, the file is opened by filename,
    and the file is seeked to the saved position.
    This means that for a successful unpickle, the original file still has to
    be accessible with its filename.

    Note
    ----
    This class only supports reading files in binary mode. If you need to open
    to open a compressed file in text mode, use :func:`bz2_pickle_open`.

    Parameters
    ----------
    name : str
        either a text or byte string giving the name (and the path
        if the file isn't in the current working directory) of the file to
        be opened.
    mode : str
        can only be 'r', 'rb' to make pickle work.

    Example
    -------
    ::

        >>> import pickle
        >>> from MDAnalysis.tests.datafiles import XYZ_bz2
        >>> file = BZ2Picklable(XYZ_bz2)
        >>> _ = file.readline()
        >>> file_pickled = pickle.loads(pickle.dumps(file))
        >>> print(file.tell(), file_pickled.tell())
        5 5

    See Also
    ---------
    FileIOPicklable
    BufferIOPicklable
    TextIOPicklable
    GzipPicklable


    .. versionadded:: 2.0.0
    """

    def __init__(self, name, mode="rb"):
        self._bz_mode = mode
        super().__init__(name, mode)

    def __getstate__(self):
        if not self._bz_mode.startswith("r"):
            raise RuntimeError(
                "Can only pickle files that were opened "
                "in read mode, not {}".format(self._bz_mode)
            )
        return {"name_val": self._fp.name, "tell_val": self.tell()}

    def __setstate__(self, args):
        name = args["name_val"]
        tell = args["tell_val"]
        self.__init__(name)
        try:
            self.seek(tell)
        except KeyError:
            pass


class GzipPicklable(gzip.GzipFile):
    """Gzip file object (read-only) that can be pickled.

    This class provides a file-like object (as returned by :func:`gzip.open`,
    namely :class:`gzip.GzipFile`) that, unlike standard Python file objects,
    can be pickled. Only read mode is supported.

    When the file is pickled, filename and position of the open file handle in
    the file are saved. On unpickling, the file is opened by filename,
    and the file is seeked to the saved position.
    This means that for a successful unpickle, the original file still has to
    be accessible with its filename.

    Note
    ----
    This class only supports reading files in binary mode. If you need to open
    to open a compressed file in text mode, use the :func:`gzip_pickle_open`.

    Parameters
    ----------
    name : str
        either a text or byte string giving the name (and the path
        if the file isn't in the current working directory) of the file to
        be opened.
    mode : str
        can only be 'r', 'rb' to make pickle work.

    Example
    -------
    ::

        >>> import pickle
        >>> from MDAnalysis.tests.datafiles import MMTF_gz
        >>> file = GzipPicklable(MMTF_gz)
        >>> _ = file.readline()
        >>> file_pickled = pickle.loads(pickle.dumps(file))
        >>> print(file.tell(), file_pickled.tell())
        1218 1218

    See Also
    ---------
    FileIOPicklable
    BufferIOPicklable
    TextIOPicklable
    BZ2Picklable


    .. versionadded:: 2.0.0
    """

    def __init__(self, name, mode="rb"):
        self._gz_mode = mode
        super().__init__(name, mode)

    def __getstate__(self):
        if not self._gz_mode.startswith("r"):
            raise RuntimeError(
                "Can only pickle files that were opened "
                "in read mode, not {}".format(self._gz_mode)
            )
        return {"name_val": self.name, "tell_val": self.tell()}

    def __setstate__(self, args):
        name = args["name_val"]
        tell = args["tell_val"]
        self.__init__(name)
        try:
            self.seek(tell)
        except KeyError:
            pass


def pickle_open(name, mode="rt"):
    """Open file and return a stream with pickle function implemented.

    This function returns a FileIOPicklable object wrapped in a
    BufferIOPicklable class when given the "rb" reading mode,
    or a FileIOPicklable object wrapped in a TextIOPicklable class with the "r"
    or "rt" reading mode. It can be used as a context manager, and replace the
    built-in :func:`open` function in read mode that only returns an
    unpicklable file object.
    In order to serialize a :class:`MDAnalysis.core.Universe`, this function
    can used to open trajectory/topology files. This object composition is more
    flexible and easier than class inheritance to implement pickling
    for new readers.

    Note
    ----
    Can be only used with read mode.

    Parameters
    ----------
    name : str
        either a text or byte string giving the name (and the path
        if the file isn't in the current working directory) of the file to
        be opened.
    mode: {'r', 'rt', 'rb'} (optional)
        'r':  open for reading in text mode;
        'rt': read in text mode (default);
        'rb': read in binary mode;

    Returns
    -------
    stream-like object: BufferIOPicklable or TextIOPicklable
        when mode is 'r' or 'rt', returns TextIOPicklable;
        when mode is 'rb', returns BufferIOPicklable

    Raises
    ------
    ValueError
        if `mode` is not one of the allowed read modes

    Examples
    -------
    open as context manager::

        with pickle_open('filename') as f:
            line = f.readline()

    open as function::

        f = pickle_open('filename')
        line = f.readline()
        f.close()

    See Also
    --------
    :func:`MDAnalysis.lib.util.anyopen`
    :func:`io.open`


    .. versionadded:: 2.0.0
    """
    if mode not in {"r", "rt", "rb"}:
        raise ValueError(
            "Only read mode ('r', 'rt', 'rb') " "files can be pickled."
        )
    name = os.fspath(name)
    raw = FileIOPicklable(name)
    if mode == "rb":
        return BufferIOPicklable(raw)
    elif mode in {"r", "rt"}:
        return TextIOPicklable(raw)


def bz2_pickle_open(name, mode="rb"):
    """Open a bzip2-compressed file in binary or text mode
    with pickle function implemented.

    This function returns a BZ2Picklable object when given the "rb" or "r"
    reading mode, or a BZ2Picklable object wrapped in a TextIOPicklable class
    with the "rt" reading mode.
    It can be used as a context manager, and replace the built-in
    :func:`bz2.open` function in read mode that only returns an
    unpicklable file object.

    Note
    ----
    Can be only used with read mode.

    Parameters
    ----------
    name : str
        either a text or byte string giving the name (and the path
        if the file isn't in the current working directory) of the file to
        be opened.
    mode: {'r', 'rt', 'rb'} (optional)
        'r':  open for reading in binary mode;
        'rt': read in text mode;
        'rb': read in binary mode; (default)

    Returns
    -------
    stream-like object: BZ2Picklable or TextIOPicklable
        when mode is 'rt', returns TextIOPicklable;
        when mode is 'r' or 'rb', returns BZ2Picklable

    Raises
    ------
    ValueError
        if `mode` is not one of the allowed read modes

    Examples
    -------
    open as context manager::

        with bz2_pickle_open('filename') as f:
            line = f.readline()

    open as function::

        f = bz2_pickle_open('filename')
        line = f.readline()
        f.close()

    See Also
    --------
    :func:`io.open`
    :func:`bz2.open`
    :func:`MDAnalysis.lib.util.anyopen`
    :func:`MDAnalysis.lib.picklable_file_io.pickle_open`
    :func:`MDAnalysis.lib.picklable_file_io.gzip_pickle_open`


    .. versionadded:: 2.0.0
    """
    if mode not in {"r", "rt", "rb"}:
        raise ValueError(
            "Only read mode ('r', 'rt', 'rb') " "files can be pickled."
        )
    bz_mode = mode.replace("t", "")
    binary_file = BZ2Picklable(name, bz_mode)
    if "t" in mode:
        return TextIOPicklable(binary_file)
    else:
        return binary_file


def gzip_pickle_open(name, mode="rb"):
    """Open a gzip-compressed file in binary or text mode
    with pickle function implemented.

    This function returns a GzipPicklable object when given the "rb" or "r"
    reading mode, or a GzipPicklable object wrapped in a TextIOPicklable class
    with the "rt" reading mode.
    It can be used as a context manager, and replace the built-in
    :func:`gzip.open` function in read mode that only returns an
    unpicklable file object.

    Note
    ----
    Can be only used with read mode.

    Parameters
    ----------
    name : str
        either a text or byte string giving the name (and the path
        if the file isn't in the current working directory) of the file to
        be opened.
    mode: {'r', 'rt', 'rb'} (optional)
        'r':  open for reading in binary mode;
        'rt': read in text mode;
        'rb': read in binary mode; (default)

    Returns
    -------
    stream-like object: GzipPicklable or TextIOPicklable
        when mode is 'rt', returns TextIOPicklable;
        when mode is 'r' or 'rb', returns GzipPicklable

    Raises
    ------
    ValueError
        if `mode` is not one of the allowed read modes

    Examples
    -------
    open as context manager::

        with gzip_pickle_open('filename') as f:
            line = f.readline()

    open as function::

        f = gzip_pickle_open('filename')
        line = f.readline()
        f.close()

    See Also
    --------
    :func:`io.open`
    :func:`gzip.open`
    :func:`MDAnalysis.lib.util.anyopen`
    :func:`MDAnalysis.lib.picklable_file_io.pickle_open`
    :func:`MDAnalysis.lib.picklable_file_io.bz2_pickle_open`


    .. versionadded:: 2.0.0
    """
    if mode not in {"r", "rt", "rb"}:
        raise ValueError(
            "Only read mode ('r', 'rt', 'rb') " "files can be pickled."
        )
    gz_mode = mode.replace("t", "")
    binary_file = GzipPicklable(name, gz_mode)
    if "t" in mode:
        return TextIOPicklable(binary_file)
    else:
        return binary_file



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

r"""Water Bridge analysis --- :mod:`MDAnalysis.analysis.hydrogenbonds.WaterBridgeAnalysis`
==========================================================================================

:Author: Zhiyi Wu
:Year: 2017-2018
:Copyright: Lesser GNU Public License v2.1+
:Maintainer: Zhiyi Wu <zhiyi.wu@gtc.ox.ac.uk>,  `@xiki-tempula`_ on GitHub


.. _`@xiki-tempula`: https://github.com/xiki-tempula


Given a :class:`~MDAnalysis.core.universe.Universe` (simulation
trajectory with 1 or more frames) measure all water bridges for each
frame between selections 1 and 2.
Water bridge is defined as a bridging water which simultaneously forms
two hydrogen bonds with atoms from both selection 1 and selection 2.

A water bridge can form between two hydrogen bond acceptors.

e.g. -CO\ :sub:`2`\ :sup:`-`:···H−O−H···:\ :sup:`-`\ O\ :sub:`2`\ C-

A water bridge can also form between two hydrogen bond donors.

e.g. -NH···:O:···HN- (where O is the oxygen of a bridging water)

A hydrogen bond acceptor and another hydrogen bond donor can be bridged by a
water.

e.g. -CO\ :sub:`2`\ :sup:`-`:···H−O:···HN- (where H−O is part of **H−O**\ −H)

A higher order water bridge is defined as more than one water bridging
hydrogen bond acceptor and donor. An example of a second order water bridge:

e.g. -CO\ :sub:`2`\ :sup:`-`:···H−O:···H−O:···HN-
(where H−O is part of **H−O**\ −H)

The following keyword arguments are important to control the behaviour of the
water bridge analysis:

- *water_selection* (``resname SOL``): the selection string for the bridging water
- *order* the maximum number of water bridging both ends
- donor-acceptor *distance* (Å): 3.0
- Angle *cutoff* (degrees): 120.0
- *forcefield* to switch between default values for different force fields
- *donors* and *acceptors* atom types (to add additional atom names)

Theory
------

This module attempts to find multi-order water bridge by an approach similar
to breadth-first search, where the first solvation shell of selection 1 is
selected, followed by the selection of the second solvation shell as well as
any hydrogen bonding partner from selection 1. After that, the third solvation
shell, as well as any binding partners from selection 2, are detected. This
process is repeated until the maximum order of water bridges is reached.

.. _wb_Analysis_Network:

Output as Network
-----------------

Since the waters connecting the two ends of the selections are by nature a
network. We provide a network representation of the water network. Water bridge
data are returned per frame, which is stored in
:attr:`WaterBridgeAnalysis.results.network`. Each frame is represented as a
dictionary, where the keys are the hydrogen bonds originating from selection
1 and the values are new dictionaries representing the hydrogen bonds coming
out of the corresponding molecules making hydrogen bonds with selection 1.

As for the hydrogen bonds which reach the selection 2, the values of the
corresponding keys are None. One example where selection 1 and selection 2 are
joined by one water molecule (A) which also hydrogen bond to another water (B)
which also hydrogen bond to selection 2 would be represented as ::

    # (selection 1)-O:···H-O(water 1):···H-(selection 2)
    #                      |             :
    #                      H·············O-H(water2)
    #                                    H
    {(sele1_acceptor, None, water1_donor, water1_donor_heavy, distance, angle):
         {(water1_acceptor, None, sele2_donor, sele2_donor_heavy,
         distance, angle): None},
         {(water1_donor, water1_donor_heavy, water2_acceptor, None,
         distance, angle):
              {(water2_acceptor, None, sele2_donor, sele2_donor_heavy,
              distance, angle): None}
          },
    }

The atoms are represented by atom index and if the atom is hydrogen bond donor,
it is followed by the index of the corresponding heavy atom
``(donor_proton, donor_heavy_atom)``.
If the atom is a hydrogen bond acceptor, it is followed by none.

.. _wb_Analysis_Timeseries:

Output as Timeseries
--------------------

For lower order water bridges, it might be desirable to represent the
connections as :attr:`WaterBridgeAnalysis.results.timeseries`. The results
are returned per frame and are a list of hydrogen bonds between the selection
1 or selection 2 and the bridging waters. Due to the complexity of the higher
order water bridge and the fact that one hydrogen bond between two waters can
appear in both third and fourth order water bridge, the hydrogen bonds in the
:attr:`WaterBridgeAnalysis.results.timeseries` attribute are generated in a
depth-first search manner to avoid duplication. Example code of how
:attr:`WaterBridgeAnalysis.results.timeseries` is generated::

    def network2timeseries(network, timeseries):
        '''Traverse the network in a depth-first fashion.
        expand_timeseries will expand the compact representation to the
        familiar timeseries representation.'''

        if network is None:
            return
        else:
            for node in network:
                timeseries.append(expand_timeseries(node))
                network2timeseries(network[node], timeseries)

    timeseries = []
    network2timeseries(network, timeseries)

An example would be. ::

    results = [
        [ # frame 1
           [ <donor index>, <acceptor index>,
            (<donor residue name>, <donor residue number>, <donor atom name>),
            (<acceptor residue name>, <acceptor residue number>,
            <acceptor atom name>),
             <distance>, <angle>],
           ....
        ],
        [ # frame 2
          [ ... ], [ ... ], ...
        ],
        ...
    ]

Using the :meth:`WaterBridgeAnalysis.generate_table` method one can reformat
the results as a flat "normalised" table that is easier to import into a
database or dataframe for further processing.

Detection of water bridges
--------------------------
Water bridges are recorded if a bridging water simultaneously forms
hydrogen bonds with selection 1 and selection 2.

Hydrogen bonds are detected based on a geometric criterion:

1. The distance between acceptor and hydrogen is less than or equal to
   `distance` (default is 3 Å).

2. The angle between donor-hydrogen-acceptor is greater than or equal to
   `angle` (default is 120º).

The cut-off values `angle` and `distance` can be set as keywords to
:class:`WaterBridgeAnalysis`.

Donor and acceptor heavy atoms are detected from atom names. The current
defaults are appropriate for the CHARMM27 and GLYCAM06 force fields as defined
in Table `Default atom names for water bridge analysis`_.

Hydrogen atoms bonded to a donor are searched based on its distance to the
donor. The algorithm searches for all hydrogens
(name "H*" or name "[123]H" or type "H") in the same residue as the donor atom
within a cut-off distance of 1.2 Å.

.. _Default atom names for water bridge analysis:

.. table:: Default heavy atom names for CHARMM27 force field.

   =========== ==============  =========== ====================================
   group       donor           acceptor    comments
   =========== ==============  =========== ====================================
   main chain  N               O, OC1, OC2 OC1, OC2 from amber99sb-ildn
                                           (Gromacs)
   water       OH2, OW         OH2, OW     SPC, TIP3P, TIP4P (CHARMM27,Gromacs)

   ARG         NE, NH1, NH2
   ASN         ND2             OD1
   ASP                         OD1, OD2
   CYS         SG
   CYH                         SG          possible false positives for CYS
   GLN         NE2             OE1
   GLU                         OE1, OE2
   HIS         ND1, NE2        ND1, NE2    presence of H determines if donor
   HSD         ND1             NE2
   HSE         NE2             ND1
   HSP         ND1, NE2
   LYS         NZ
   MET                         SD          see e.g. :footcite:p:`Gregoret1991`
   SER         OG              OG
   THR         OG1             OG1
   TRP         NE1
   TYR         OH              OH
   =========== ==============  =========== ====================================

.. table:: Heavy atom types for GLYCAM06 force field.

   =========== =========== ==================
   element     donor       acceptor
   =========== =========== ==================
   N           N,NT,N3     N,NT
   O           OH,OW       O,O2,OH,OS,OW,OY
   S                       SM
   =========== =========== ==================

Donor and acceptor names for the CHARMM27 force field will also work for e.g.
OPLS/AA or amber (tested in Gromacs). Residue names in the table are for
information only and are not taken into account when determining acceptors and
donors. This can potentially lead to some ambiguity in the assignment of
donors/acceptors for residues such as histidine or cytosine.

For more information about the naming convention in GLYCAM06 have a look at the
`Carbohydrate Naming Convention in Glycam`_.

.. _`Carbohydrate Naming Convention in Glycam`:
   http://glycam.ccrc.uga.edu/documents/FutureNomenclature.htm

The lists of donor and acceptor names can be extended by providing lists of
atom names in the `donors` and `acceptors` keywords to
:class:`WaterBridgeAnalysis`. If the lists are entirely inappropriate
(e.g. when analysing simulations done with a force field that uses very
different atom names) then one should either use the value "other" for
`forcefield` to set no default values or derive a new class and set the
default list oneself::

 class WaterBridgeAnalysis_OtherFF(WaterBridgeAnalysis):
       DEFAULT_DONORS = {"OtherFF": tuple(set([...]))}
       DEFAULT_ACCEPTORS = {"OtherFF": tuple(set([...]))}

Then simply use the new class instead of the parent class and call it with
```forcefield` = "OtherFF"``. Please also consider contributing the list of
heavy atom names to MDAnalysis.


.. rubric:: References

.. footbibliography::


How to perform ``WaterBridgeAnalysis``
--------------------------------------

All water bridges between arginine and aspartic acid can be analysed with ::

  import MDAnalysis
  from MDAnalysis.analysis.hydrogenbonds import WaterBridgeAnalysis

  u = MDAnalysis.Universe('topology', 'trajectory')
  w = WaterBridgeAnalysis(u, 'resname ARG', 'resname ASP')
  w.run()

The maximum number of bridging waters detected can be changed using the order
keyword. ::

  w = WaterBridgeAnalysis(u, 'resname ARG', 'resname ASP', order=3)

Thus, a maximum of three bridging waters will be detected.

An example of using the :attr:`~WaterBridgeAnalysis` would be
detecting the percentage of time a certain water bridge exits.

Trajectory :code:`u` has two frames, where the first frame contains a water
bridge from the oxygen of the first arginine to one of the oxygens in the
carboxylic group of aspartate (ASP3:OD1). In the second frame, the same water
bridge forms but is between the oxygen of the arginine and the other oxygen in
the carboxylic group (ASP3:OD2). ::

  # index residue id residue name atom name
  #     0          1          ARG         O
  #     1          2          SOL        OW
  #     2          2          SOL       HW1
  #     3          2          SOL       HW2
  #     4          3          ASP       OD1
  #     5          3          ASP       OD2
  print(w.results.timeseries)

prints out ::

  [ # frame 1
    # A water bridge SOL2 links O from ARG1 to the carboxylic group OD1 of ASP3
   [[0,2,('ARG',1,  'O'),('SOL',2,'HW1'),  3.0,180],
    [3,4,('SOL',2,'HW2'),('ASP',3,'OD1'),  3.0,180],
   ],
    # frame 2
    # Another water bridge SOL2 links O from ARG1 to the other oxygen of the
    # carboxylic group OD2 of ASP3
   [[0,2,('ARG',1,  'O'),('SOL',2,'HW1'),  3.0,180],
    [3,5,('SOL',2,'HW2'),('ASP',3,'OD2'),  3.0,180],
   ],
  ]


.. _wb_count_by_type:

Use ``count_by_type``
---------------------

We can use the :meth:`~WaterBridgeAnalysis.count_by_type` to
generate the frequency of all water bridges in the simulation. ::

  w.count_by_type()

Returns ::

  [(0, 3, 'ARG', 1, 'O', 'ASP', 3, 'OD1', 0.5),
   (0, 4, 'ARG', 1, 'O', 'ASP', 3, 'OD2', 0.5),]

You might think that the OD1 and OD2 are the same oxygen and the aspartate has
just flipped and thus, they should be counted as the same type of water bridge.
The type of the water bridge can be customised by supplying an analysis
function to :meth:`~WaterBridgeAnalysis.count_by_type`.

The analysis function has two parameters. The current and the output. The
current is a list of hydrogen bonds from selection 1 to selection 2, formatted
in the same fashion as :attr:`WaterBridgeAnalysis.network`, and an example will
be ::

  [
  # sele1 acceptor idx,   , water donor index, donor heavy atom idx, dist, ang.
   [                 0, None,                 2,                   1, 3.0,180],
  # water donor idx, donor heavy atom idx, sele2 acceptor idx, distance, angle.
   [              3,                    1,                  4, None, 3.0,180],]

Where ``current[0]`` is the first hydrogen bond originating from selection 1
and ``current[-1]`` is the final hydrogen bond ending in selection 2. The
output sums up all the information in the current frame and is a dictionary
with a user-defined key and the value is the weight of the corresponding key.
During the analysis phase, the function analysis iterates through all the water
bridges and modify the output in-place. At the end of the analysis, the keys
from all the frames are collected and the corresponding values will be summed
up and returned. ::

  def analysis(current, output, u):
      r'''This function defines how the type of water bridge should be
      specified.

        Parameters
        ----------
        current : list
            The current water bridge being analysed is a list of hydrogen bonds
            from selection 1 to selection 2.
        output : dict
            A dictionary which is modified in-place where the key is the type
            of the water bridge and the value is the weight of this type of
            water bridge.
        u : MDAnalysis.universe
            The current Universe for looking up atoms.'''

      # decompose the first hydrogen bond.
      sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle =
      current[0]
      # decompose the last hydrogen bond.
      atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle =
      current[-1]
      # expand the atom index to the resname, resid, atom names
      sele1 = u.atoms[sele1_index]
      sele2 = u.atoms[sele2_index]
      (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid,
      sele1.name)
      (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid,
      sele2.name)
      # if the residue name is ASP and the atom name is OD2 or OD1,
      # the atom name is changed to OD
      if s2_resname == 'ASP' and (s2_name == 'OD1' or s2_name == 'OD2'):
          s2_name = 'OD'
      # setting up the key which defines this type of water bridge.
      key = (s1_resname, s1_resid, s1_name, s2_resname, s2_resid, s2_name)
      # The number of this type of water bridge is incremented by 1.
      output[key] += 1

  w.count_by_type(analysis_func=analysis)

Returns ::

  [(('ARG', 1, 'O', 'ASP', 3, 'OD'), 1.0),]

Note that the result is arranged in the format of
``(key, the proportion of time)``. When no custom analysis function is supplied
, the key is expanded and is formatted as ::

  [('ARG', 1, 'O', 'ASP', 3, 'OD', 1.0),]

Some people might only interested in contacts between residues and pay no
attention to the details regarding the atom name. However, since multiple water
bridges can exist between two residues, which sometimes can give a result such
that the water bridge between two residues exists 300% of the time. Though this
might be a desirable result for some people, others might want the water bridge
between two residues to be only counted once per frame. This can also be
achieved by supplying an analysis function to
:meth:`~WaterBridgeAnalysis.count_by_type`. ::

  def analysis(current, output, u):
      '''This function defines how the type of water bridge should be specified
      .

        Parameters
        ----------
        current : list
            The current water bridge being analysed is a list of hydrogen bonds
            from selection 1 to selection 2.
        output : dict
            A dictionary which is modified in-place where the key is the type
            of the water bridge and the value is the weight of this type of
            water bridge.
        u : MDAnalysis.universe
            The current Universe for looking up atoms.
      '''

      # decompose the first hydrogen bond.
      sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle =
      current[0]
      # decompose the last hydrogen bond.
      atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle =
      current[-1]
      # expand the atom index to the resname, resid, atom names
      sele1 = u.atoms[sele1_index]
      sele2 = u.atoms[sele2_index]
      (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid,
      sele1.name)
      (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid,
      sele2.name)
      # s1_name and s2_name are not included in the key
      key = (s1_resname, s1_resid, s2_resname, s2_resid)

      # Each residue is only counted once per frame
      output[key] = 1

  w.count_by_type(analysis_func=analysis)

Returns ::

  [(('ARG', 1, 'ASP', 3), 1.0),]

On the other hand, other people may insist that the first order and
second-order water bridges shouldn't be mixed together, which can also be
achieved by supplying an analysis function to
:meth:`~WaterBridgeAnalysis.count_by_type`.  ::

  def analysis(current, output, u):
      '''This function defines how the type of water bridge should be specified
      .

        Parameters
        ----------
        current : list
            The current water bridge being analysed is a list of hydrogen bonds
            from selection 1 to selection 2.
        output : dict
            A dictionary which is modified in-place where the key is the type
            of the water bridge and the value is the weight of this type of
            water bridge.
        u : MDAnalysis.universe
            The current Universe for looking up atoms.
      '''

      # decompose the first hydrogen bond.
      sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle =
      current[0]
      # decompose the last hydrogen bond.
      atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle =
      current[-1]
      # expand the atom index to the resname, resid, atom names
      sele1 = u.atoms[sele1_index]
      sele2 = u.atoms[sele2_index]
      (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid,
      sele1.name)
      (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid,
      sele2.name)
      # order of the current water bridge is computed
      order_of_water_bridge = len(current) - 1
      # and is included in the key
      key = (s1_resname, s1_resid, s2_resname, s2_resid, order_of_water_bridge)
      # The number of this type of water bridge is incremented by 1.
      output[key] += 1

  w.count_by_type(analysis_func=analysis)

The extra number 1 precede the 1.0 indicate that this is a first order water
bridge ::

  [(('ARG', 1, 'ASP', 3, 1), 1.0),]

Some people might not be interested in the interactions related to arginine.
The undesirable interactions can be discarded by supplying an analysis function
to :meth:`~WaterBridgeAnalysis.count_by_type`.  ::

  def analysis(current, output, u):
      '''This function defines how the type of water bridge should be
      specified.

        Parameters
        ----------
        current : list
            The current water bridge being analysed is a list of hydrogen bonds
            from selection 1 to selection 2.
        output : dict
            A dictionary which is modified in-place where the key is the type
            of the water bridge and the value is the number of this type of
            water bridge.
        u : MDAnalysis.universe
            The current Universe for looking up atoms.
      '''

      # decompose the first hydrogen bond.
      sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle =
      current[0]
      # decompose the last hydrogen bond.
      atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle =
      current[-1]
      # expand the atom index to the resname, resid, atom names
      sele1 = u.atoms[sele1_index]
      sele2 = u.atoms[sele2_index]
      (s1_resname, s1_resid, s1_name) = (sele1.resname, sele1.resid,
      sele1.name)
      (s2_resname, s2_resid, s2_name) = (sele2.resname, sele2.resid,
      sele2.name)
      if not s1_resname == 'ARG':
          key = (s1_resname, s1_resid, s2_resname, s2_resid)
          output[key] += 1

  w.count_by_type(analysis_func=analysis)

Returns nothing in this case ::

  [,]

Additional keywords can be supplied to the analysis function by passing through
:meth:`~WaterBridgeAnalysis.count_by_type`.  ::

  def analysis(current, output, **kwargs):
      ...
  w.count_by_type(analysis_func=analysis, **kwargs)


.. _wb_count_by_time:

Use ``count_by_time``
---------------------

:meth:`~WaterBridgeAnalysis.count_by_type` aggregates data across frames, which
might be desirable in some cases but not the others.
:meth:`~WaterBridgeAnalysis.count_by_time` provides additional functionality
for aggregating results for each frame.

The default behaviour of :meth:`~WaterBridgeAnalysis.count_by_time` is counting
the number of water bridges from selection 1 to selection 2 for each frame.
Take the previous ASP, ARG salt bridge for example:  ::

  w.count_by_time()

As one water bridge is found in both frames, the method returns ::

  [(1.0, 1), (2.0, 1), ]

Similar to :meth:`~WaterBridgeAnalysis.count_by_type`
The behaviour of :meth:`~WaterBridgeAnalysis.count_by_time` can also be
modified by supplying an analysis function.

Suppose we want to count

  - the **number** of water molecules involved in bridging selection 1 to
    selection 2.
  - only if water bridge terminates in atom name **OD1 of ASP**.
  - only when water bridge is joined by less than **two** water.

The analysis function can be written as::

  def analysis(current, output, u, **kwargs):
      '''This function defines how the counting of water bridge should be
      specified.

        Parameters
        ----------
        current : list
            The current water bridge being analysed is a list of hydrogen bonds
            from selection 1 to selection 2.
        output : dict
            A dictionary which is modified in-place where the key is the type
            of the water bridge and the value is the number of this type of
            water  bridge.
        u : MDAnalysis.universe
            The current Universe for looking up atoms.
      '''

      # decompose the first hydrogen bond.
      sele1_index, sele1_heavy_index, atom2, heavy_atom2, dist, angle =
      current[0]
      # decompose the last hydrogen bond.
      atom1, heavy_atom1, sele2_index, sele2_heavy_index, dist, angle =
      current[-1]
      # expand the atom index to the resname, resid, atom names
      sele1 = u.atoms[sele1_index]
      sele2 = u.atoms[sele2_index]
      (s1_resname, s1_resid, s1_name) =
      (sele1.resname, sele1.resid, sele1.name)
      (s2_resname, s2_resid, s2_name) =
      (sele2.resname, sele2.resid, sele2.name)

      # only the residue name is ASP and the atom name is OD1,
      if s2_resname == 'ASP' and s2_name == 'OD1':
          # only if the order of water bridge is less than 2
          if len(current) -1 < 2:
              # extract all water molecules involved in the water bridge
              # extract the first water from selection 1
              s1_index, to_index, (s1_resname, s1_resid, s1_name),
              (to_resname, to_resid, to_name), dist, angle = current[0]
              key = (to_resname, to_resid)
              output[key] = 1

              # extract all the waters between selection 1 and selection 2
              for hbond in current[1:-1]:
                  # decompose the hydrogen bond.
                  from_index, to_index, (from_resname, from_resid, from_name),
                  (to_resname, to_resid, to_name), dist, angle = hbond
                  # add first water
                  key1 = (from_resname, from_resid)
                  output[key1] = 1
                  # add second water
                  key2 = (to_resname, to_resid)
                  output[key2] = 1

              # extract the last water to selection 2
              from_index, s2_index, (from_resname, from_resid, from_name),
              (s2_resname, s2_resid, s2_name), dist, angle = current[-1]
              key = (from_resname, from_resid)
              output[key] = 1

  w.count_by_time(analysis_func=analysis)

Returns ::

  [(1.0, 1), (2.0, 0),]

Classes
-------

.. autoclass:: WaterBridgeAnalysis
   :members:

   .. attribute:: timesteps

      List of the times of each timestep. This can be used together with
      :attr:`~WaterBridgeAnalysis.results.timeseries` to find the specific
      time point of a water bridge existence.

   .. attribute:: results.network

      Network representation of the water network.

      .. versionadded:: 2.0.0

   .. attribute:: network

      Alias to the :attr:`results.network` attribute.

      .. deprecated:: 2.0.0
         Will be removed in MDAnalysis 3.0.0. Please use
         :attr:`results.network` instead.

   .. attribute:: table

      .. deprecated:: 2.0.0
         Will be removed in MDAnalysis 3.0.0. Please generate
         the table with :meth:`generate_table` instead.

   .. attribute:: results.timeseries

      List of hydrogen bonds between the selection 1 or selection 2
      and the bridging waters, for each frame.

      .. versionadded:: 2.0.0

   .. attribute:: timeseries

      Alias to the :attr:`results.timeseries` attribute.

      .. deprecated:: 2.0.0
         Will be removed in MDAnalysis 3.0.0. Please use
         :attr:`results.timeseries` instead.
"""
import logging
import warnings
from collections import defaultdict

import numpy as np

from MDAnalysis import MissingDataWarning, NoDataError, SelectionError
from MDAnalysis.lib.distances import calc_angles, capped_distance
from MDAnalysis.lib.NeighborSearch import AtomNeighborSearch

from ..base import AnalysisBase

logger = logging.getLogger("MDAnalysis.analysis.WaterBridgeAnalysis")


class WaterBridgeAnalysis(AnalysisBase):
    """Perform a water bridge analysis

    The analysis of the trajectory is performed with the
    :meth:`WaterBridgeAnalysis.run` method. The result is stored in
    :attr:`WaterBridgeAnalysis.results.timeseries`. See
    :meth:`~WaterBridgeAnalysis.run` for the format.

    .. versionadded:: 0.17.0

    """

    # use tuple(set()) here so that one can just copy&paste names from the
    # table; set() takes care for removing duplicates. At the end the
    # DEFAULT_DONORS and DEFAULT_ACCEPTORS should simply be tuples.

    #: default heavy atom names whose hydrogens are treated as *donors*
    #: (see :ref:`Default atom names for water bridge analysis`);
    #: use the keyword `donors` to add a list of additional donor names.
    DEFAULT_DONORS = {
        "CHARMM27": tuple(
            {
                "N",
                "OH2",
                "OW",
                "NE",
                "NH1",
                "NH2",
                "ND2",
                "SG",
                "NE2",
                "ND1",
                "NZ",
                "OG",
                "OG1",
                "NE1",
                "OH",
            }
        ),
        "GLYCAM06": tuple({"N", "NT", "N3", "OH", "OW"}),
        "other": tuple(set([])),
    }

    #: default atom names that are treated as hydrogen *acceptors*
    #: (see :ref:`Default atom names for water bridge analysis`);
    #: use the keyword `acceptors` to add a list of additional acceptor names.
    DEFAULT_ACCEPTORS = {
        "CHARMM27": tuple(
            {
                "O",
                "OC1",
                "OC2",
                "OH2",
                "OW",
                "OD1",
                "OD2",
                "SG",
                "OE1",
                "OE1",
                "OE2",
                "ND1",
                "NE2",
                "SD",
                "OG",
                "OG1",
                "OH",
            }
        ),
        "GLYCAM06": tuple(
            {"N", "NT", "O", "O2", "OH", "OS", "OW", "OY", "SM"}
        ),
        "other": tuple(set([])),
    }

    #: A :class:`collections.defaultdict` of covalent radii of common donors
    #: (used in :meth`_get_bonded_hydrogens_list` to check if a hydrogen is
    #: sufficiently close to its donor heavy atom). Values are stored for
    #: N, O, P, and S. Any other heavy atoms are assumed to have hydrogens
    #: covalently bound at a maximum distance of 1.5 Å.
    r_cov = defaultdict(
        lambda: 1.5, N=1.31, O=1.31, P=1.58, S=1.55  # default value
    )  # noqa: E741

    def __init__(
        self,
        universe,
        selection1="protein",
        selection2="not resname SOL",
        water_selection="resname SOL",
        order=1,
        selection1_type="both",
        update_selection=False,
        update_water_selection=True,
        filter_first=True,
        distance_type="hydrogen",
        distance=3.0,
        angle=120.0,
        forcefield="CHARMM27",
        donors=None,
        acceptors=None,
        output_format="sele1_sele2",
        debug=None,
        pbc=False,
        **kwargs,
    ):
        """Set up the calculation of water bridges between two selections in a
        universe.

        The timeseries is accessible as the attribute
        :attr:`WaterBridgeAnalysis.results.timeseries`.

        If no hydrogen bonds are detected or if the initial check fails, look
        at the log output (enable with :func:`MDAnalysis.start_logging` and set
        `verbose` ``=True``). It is likely that the default names for donors
        and acceptors are not suitable (especially for non-standard
        ligands). In this case, either change the `forcefield` or use
        customized `donors` and/or `acceptors`.

        Parameters
        ----------
        universe : Universe
            Universe object
        selection1 : str (optional)
            Selection string for first selection ['protein']
        selection2 : str (optional)
            Selection string for second selection ['not resname SOL']
            This string selects everything except water where water is assumed
            to have a residue name as SOL.
        water_selection : str (optional)
            Selection string for bridging water selection ['resname SOL']
            The default selection assumes that the water molecules have residue
            name "SOL". Change it to the appropriate selection for your
            specific force field.

            However, in theory, this selection can be anything which forms
            a hydrogen bond with selection 1 and selection 2.
        order : int (optional)
            The maximum number of water bridges linking both selections.
            if the order is set to 3, then all the residues linked with less
            than three water molecules will be detected. [1]

            Computation of high order water bridges can be very time-consuming.
            Think carefully before running the calculation, do you really want
            to compute the 20th order water bridge between domain A and domain
            B or you just want to know the third order water bridge between two
            residues.
        selection1_type : {"donor", "acceptor", "both"} (optional)
            Selection 1 can be 'donor', 'acceptor' or 'both'. Note that the
            value for `selection1_type` automatically determines how
            `selection2` handles donors and acceptors: If `selection1` contains
            'both' then `selection2` will also contain 'both'. If `selection1`
            is set to 'donor' then `selection2` is 'acceptor' (and vice versa).
            ['both'].
        update_selection : bool (optional)
            Update selection 1 and 2 at each frame. Setting to ``True`` if the
            selection is not static. Selections are filtered first to speed up
            performance. Thus, setting to ``True`` is recommended if contact
            surface between selection 1 and selection 2 is constantly
            changing. [``False``]
        update_water_selection : bool (optional)
            Update selection of water at each frame. Setting to ``False`` is
            **only** recommended when the total amount of water molecules in
            the simulation are small and when water molecules remain static
            across the simulation.

            However, in normal simulations, only a tiny proportion of water is
            engaged in the formation of water bridge. It is recommended to
            update the water selection and set keyword `filter_first` to
            ``True`` so as to filter out water not residing between the two
            selections. [``True``]
        filter_first : bool (optional)
            Filter the water selection to only include water within 4 Å +
            `order` * (2 Å + `distance`) away from `both` selection 1 and
            selection 2.
            Selection 1 and selection 2 are both filtered to only include atoms
            with the same distance away from the other selection. [``True``]
        distance : float (optional)
            Distance cutoff for hydrogen bonds; only interactions with a H-A
            distance <= `distance` (and the appropriate D-H-A angle, see
            `angle`) are recorded. (Note: `distance_type` can change this to
            the D-A distance.) [3.0]
        angle : float (optional)
            Angle cutoff for hydrogen bonds; an ideal H-bond has an angle of
            180º.  A hydrogen bond is only recorded if the D-H-A angle is
            >=  `angle`. The default of 120º also finds fairly non-specific
            hydrogen interactions and possibly better value is 150º. [120.0]
        forcefield : {"CHARMM27", "GLYCAM06", "other"} (optional)
            Name of the forcefield used. Switches between different
            :attr:`~DEFAULT_DONORS` and
            :attr:`~DEFAULT_ACCEPTORS` values.
            ["CHARMM27"]
        donors : sequence (optional)
            Extra H donor atom types (in addition to those in :attr:`~DEFAULT_DONORS`).
            This shall be the name of the heavy atom that is bonded to the hydrogen.
            For example, the oxygen ('O') in the hydroxyl group. Must be a sequence.
        acceptors : sequence (optional)
            Extra H acceptor atom types (in addition to those in
            :attr:`~DEFAULT_ACCEPTORS`), must be a
            sequence.
        distance_type : {"hydrogen", "heavy"} (optional)
            Measure hydrogen bond lengths between donor and acceptor heavy
            atoms ("heavy") or between donor hydrogen and acceptor heavy
            atom ("hydrogen"). If using "heavy" then one should set the
            *distance* cutoff to a higher value such as 3.5 Å. ["hydrogen"]
        output_format: {"sele1_sele2", "donor_acceptor"} (optional)
            Setting the output format for timeseries and table. If set to
            "sele1_sele2", for each hydrogen bond, the one close to selection 1
            will be placed before selection 2. If set to "donor_acceptor", the
            donor will be placed before acceptor. "sele1_sele2"]
        debug : bool (optional)
            If set to ``True`` enables per-frame debug logging. This is
            disabled by default because it generates a very large amount of
            output in the log file. (Note that a logger must have been started
            to see the output, e.g. using :func:`MDAnalysis.start_logging`.)
        verbose : bool (optional)
            Toggle progress output. (Can also be given as keyword argument to
            :meth:`run`.)

        Notes
        -----
        In order to speed up processing, atoms are filtered by a coarse
        distance criterion before a detailed hydrogen bonding analysis is
        performed (`filter_first` = ``True``).

        If selection 1 and selection 2 are very mobile during the simulation
        and the contact surface is constantly changing (i.e. residues are
        moving farther than 4 Å + `order` * (2 Å + `distance`)), you might
        consider setting the `update_selection` keywords to ``True``
        to ensure correctness.

        .. versionchanged 0.20.0
           The :attr:`WaterBridgeAnalysis.timeseries` has been updated
           see :attr:`WaterBridgeAnalysis.timeseries` for detail.
           This class is now based on
           :class:`~MDAnalysis.analysis.base.AnalysisBase`.


        """
        super(WaterBridgeAnalysis, self).__init__(
            universe.trajectory, **kwargs
        )
        self.water_selection = water_selection
        self.update_water_selection = update_water_selection
        # per-frame debugging output?
        self.debug = debug

        # set the output format
        self.output_format = output_format

        self.u = universe
        self.selection1 = selection1
        self.selection2 = selection2
        self.selection1_type = selection1_type
        if "selection2_type" in kwargs:
            raise ValueError("`selection2_type` is not a keyword argument.")

        # if the selection 1 and selection 2 are the same
        if selection1 == selection2:
            # eliminate the duplication
            self.selection1_type = "donor"
        self.update_selection = update_selection
        self.filter_first = filter_first
        self.distance = distance
        if distance_type not in {"hydrogen", "heavy"}:
            raise ValueError(
                f"Only 'hydrogen' and 'heavy' are allowed for option `distance_type' ({distance_type})."
            )
        self.distance_type = distance_type
        # will give the default behavior
        self.angle = angle
        self.pbc = pbc and all(self.u.dimensions[:3])
        self.order = order

        # set up the donors/acceptors lists
        if donors is None:
            donors = ()
        if acceptors is None:
            acceptors = ()
        self.forcefield = forcefield
        self.donors = tuple(set(self.DEFAULT_DONORS[forcefield]).union(donors))
        self.acceptors = tuple(
            set(self.DEFAULT_ACCEPTORS[forcefield]).union(acceptors)
        )

        if self.selection1_type not in ("both", "donor", "acceptor"):
            raise ValueError(
                "WaterBridgeAnalysis: "
                "Invalid selection type {0!s}".format(self.selection1_type)
            )

        # final result accessed as self.results.network
        self.results.network = []
        self.results.timeseries = None
        self.timesteps = None  # time for each frame

        self._log_parameters()

    def _log_parameters(self):
        """Log important parameters to the logfile."""
        logger.info(
            "WaterBridgeAnalysis: selection = %r (update: %r)",
            self.selection2,
            self.update_selection,
        )
        logger.info(
            "WaterBridgeAnalysis: water selection = %r (update: %r)",
            self.water_selection,
            self.update_water_selection,
        )
        logger.info(
            "WaterBridgeAnalysis: criterion: donor %s atom and "
            "acceptor atom distance <= %.3f A",
            self.distance_type,
            self.distance,
        )
        logger.info(
            "WaterBridgeAnalysis: criterion: " "angle D-H-A >= %.3f degrees",
            self.angle,
        )
        logger.info(
            "WaterBridgeAnalysis: force field %s to guess donor and \
        acceptor names",
            self.forcefield,
        )

    def _build_residue_dict(self, selection):
        # Build the residue_dict where the key is the residue name
        # The content is a dictionary where hydrogen bond donor heavy atom
        # names is the key
        # The content is the hydrogen bond donor hydrogen atom names
        atom_group = self.u.select_atoms(selection)
        for residue in atom_group.residues:
            if residue.resname not in self._residue_dict:
                self._residue_dict[residue.resname] = defaultdict(set)
            for atom in residue.atoms:
                if atom.name in self.donors:
                    self._residue_dict[residue.resname][atom.name].update(
                        self._get_bonded_hydrogens(atom).names
                    )

    def _update_donor_h(self, atom_ix, h_donors, donors_h):
        atom = self.u.atoms[atom_ix]
        residue = atom.residue
        hydrogen_names = self._residue_dict[residue.resname][atom.name]
        if hydrogen_names:
            hydrogens = residue.atoms.select_atoms(
                "name {0}".format(" ".join(hydrogen_names))
            ).ix
            for atom in hydrogens:
                h_donors[atom] = atom_ix
                donors_h[atom_ix].append(atom)

    def _update_selection(self):
        self._s1_donors = []
        self._s1_h_donors = {}
        self._s1_donors_h = defaultdict(list)
        self._s1_acceptors = []

        self._s2_donors = []
        self._s2_h_donors = {}
        self._s2_donors_h = defaultdict(list)
        self._s2_acceptors = []

        self._s1 = self.u.select_atoms(self.selection1).ix
        self._s2 = self.u.select_atoms(self.selection2).ix

        if self.filter_first and len(self._s1):
            self.logger_debug(
                "Size of selection 1 before filtering:"
                " {} atoms".format(len(self._s1))
            )
            ns_selection_1 = AtomNeighborSearch(
                self.u.atoms[self._s1], box=self.box
            )
            self._s1 = ns_selection_1.search(
                self.u.atoms[self._s2], self.selection_distance
            ).ix
        self.logger_debug(
            "Size of selection 1: {0} atoms".format(len(self._s1))
        )

        if len(self._s1) == 0:
            logger.warning(
                'Selection 1 "{0}" did not select any atoms.'.format(
                    str(self.selection1)[:80]
                )
            )
            return

        if self.filter_first and len(self._s2):
            self.logger_debug(
                "Size of selection 2 before filtering:"
                " {} atoms".format(len(self._s2))
            )
            ns_selection_2 = AtomNeighborSearch(
                self.u.atoms[self._s2], box=self.box
            )
            self._s2 = ns_selection_2.search(
                self.u.atoms[self._s1], self.selection_distance
            ).ix
        self.logger_debug(
            "Size of selection 2: {0} atoms".format(len(self._s2))
        )

        if len(self._s2) == 0:
            logger.warning(
                'Selection 2 "{0}" did not select any atoms.'.format(
                    str(self.selection2)[:80]
                )
            )
            return

        if self.selection1_type in ("donor", "both"):
            self._s1_donors = (
                self.u.atoms[self._s1]
                .select_atoms("name {0}".format(" ".join(self.donors)))
                .ix
            )
            for atom_ix in self._s1_donors:
                self._update_donor_h(
                    atom_ix, self._s1_h_donors, self._s1_donors_h
                )
            self.logger_debug(
                "Selection 1 donors: {0}".format(len(self._s1_donors))
            )
            self.logger_debug(
                "Selection 1 donor hydrogens: {0}".format(
                    len(self._s1_h_donors)
                )
            )
        if self.selection1_type in ("acceptor", "both"):
            self._s1_acceptors = (
                self.u.atoms[self._s1]
                .select_atoms("name {0}".format(" ".join(self.acceptors)))
                .ix
            )
            self.logger_debug(
                "Selection 1 acceptors: {0}".format(len(self._s1_acceptors))
            )

        if len(self._s2) == 0:
            return None
        if self.selection1_type in ("donor", "both"):
            self._s2_acceptors = (
                self.u.atoms[self._s2]
                .select_atoms("name {0}".format(" ".join(self.acceptors)))
                .ix
            )
            self.logger_debug(
                "Selection 2 acceptors: {0:d}".format(len(self._s2_acceptors))
            )
        if self.selection1_type in ("acceptor", "both"):
            self._s2_donors = (
                self.u.atoms[self._s2]
                .select_atoms("name {0}".format(" ".join(self.donors)))
                .ix
            )
            for atom_ix in self._s2_donors:
                self._update_donor_h(
                    atom_ix, self._s2_h_donors, self._s2_donors_h
                )
            self.logger_debug(
                "Selection 2 donors: {0:d}".format(len(self._s2_donors))
            )
            self.logger_debug(
                "Selection 2 donor hydrogens: {0:d}".format(
                    len(self._s2_h_donors)
                )
            )

    def _update_water_selection(self):
        self._water_donors = []
        self._water_h_donors = {}
        self._water_donors_h = defaultdict(list)
        self._water_acceptors = []

        self._water = self.u.select_atoms(self.water_selection).ix
        self.logger_debug(
            "Size of water selection before filtering:"
            " {} atoms".format(len(self._water))
        )
        if len(self._water) and self.filter_first:
            filtered_s1 = AtomNeighborSearch(
                self.u.atoms[self._water], box=self.box
            ).search(self.u.atoms[self._s1], self.selection_distance)
            if filtered_s1:
                self._water = (
                    AtomNeighborSearch(filtered_s1, box=self.box)
                    .search(self.u.atoms[self._s2], self.selection_distance)
                    .ix
                )

        self.logger_debug(
            "Size of water selection: {0} atoms".format(len(self._water))
        )

        if len(self._water) == 0:
            logger.warning(
                "Water selection '{0}' did not select any atoms.".format(
                    str(self.water_selection)[:80]
                )
            )
        else:
            self._water_donors = (
                self.u.atoms[self._water]
                .select_atoms("name {0}".format(" ".join(self.donors)))
                .ix
            )
            for atom_ix in self._water_donors:
                self._update_donor_h(
                    atom_ix, self._water_h_donors, self._water_donors_h
                )
            self.logger_debug(
                "Water donors: {0}".format(len(self._water_donors))
            )
            self.logger_debug(
                "Water donor hydrogens: {0}".format(len(self._water_h_donors))
            )
            self._water_acceptors = (
                self.u.atoms[self._water]
                .select_atoms("name {0}".format(" ".join(self.acceptors)))
                .ix
            )
            self.logger_debug(
                "Water acceptors: {0}".format(len(self._water_acceptors))
            )

    def _get_bonded_hydrogens(self, atom):
        """Find hydrogens bonded within cutoff to `atom`.

        Hydrogens are detected by either name ("H*", "[123]H*") or type ("H");
        this is not fool-proof as the atom type is not always a character but
        the name pattern should catch most typical occurrences.

        The distance from `atom` is calculated for all hydrogens in the residue
        and only those within a cutoff are kept. The cutoff depends on the
        heavy atom (more precisely, on its element, which is taken as the first
        letter of its name ``atom.name[0]``) and is parameterized in
        :attr:`WaterBridgeAnalysis.r_cov`. If no match is found then the
        default of 1.5 Å is used.


        Parameters
        ----------
        atom : groups.Atom
             heavy atom

        Returns
        -------
        hydrogen_atoms : AtomGroup or []
            list of hydrogens (can be a
            :class:`~MDAnalysis.core.groups.AtomGroup`)
            or empty list ``[]`` if none were found.
        """
        try:
            return atom.residue.atoms.select_atoms(
                "(name H* 1H* 2H* 3H* or type H) and around {0:f} name {1!s}"
                "".format(self.r_cov[atom.name[0]], atom.name)
            )
        except NoDataError:
            return []

    def logger_debug(self, *args):
        if self.debug:
            logger.debug(*args)

    def _prepare(self):
        # The distance for selection is defined as twice the maximum bond
        # length of an O-H bond (2A) plus order of water bridge times the
        # length of OH bond plus hydrogne bond distance
        self.selection_distance = 2 * 2 + self.order * (2 + self.distance)

        self.box = self.u.dimensions if self.pbc else None
        self._residue_dict = {}
        self._build_residue_dict(self.selection1)
        self._build_residue_dict(self.selection2)
        self._build_residue_dict(self.water_selection)

        self._update_selection()

        self.timesteps = []
        if len(self._s1) and len(self._s2):
            self._update_water_selection()
        else:
            logger.info(
                "WaterBridgeAnalysis: " "no atoms found in the selection."
            )

        logger.info("WaterBridgeAnalysis: initial checks passed.")

        logger.info("WaterBridgeAnalysis: starting")
        logger.debug("WaterBridgeAnalysis: donors    %r", self.donors)
        logger.debug("WaterBridgeAnalysis: acceptors %r", self.acceptors)
        logger.debug(
            "WaterBridgeAnalysis: water bridge %r", self.water_selection
        )

        if self.debug:
            logger.debug("Toggling debug to %r", self.debug)
        else:
            logger.debug(
                "WaterBridgeAnalysis: For full step-by-step "
                "debugging output use debug=True"
            )

        logger.info(
            "Starting analysis " "(frame index start=%d stop=%d, step=%d)",
            self.start,
            self.stop,
            self.step,
        )

    def _donor2acceptor(self, donors, h_donors, acceptor):
        if len(donors) == 0 or len(acceptor) == 0:
            return []
        if self.distance_type != "heavy":
            donors_idx = list(h_donors.keys())
        else:
            donors_idx = list(donors.keys())
        result = []
        # Code modified from p-j-smith
        pairs, distances = capped_distance(
            self.u.atoms[donors_idx].positions,
            self.u.atoms[acceptor].positions,
            max_cutoff=self.distance,
            box=self.box,
            return_distances=True,
        )
        if self.distance_type == "hydrogen":
            tmp_distances = distances
            tmp_donors = [h_donors[donors_idx[idx]] for idx in pairs[:, 0]]
            tmp_hydrogens = [donors_idx[idx] for idx in pairs[:, 0]]
            tmp_acceptors = [acceptor[idx] for idx in pairs[:, 1]]
        else:
            # To make sure that for the same index i, the donor (tmp_donors[i]),
            # hydrogen (tmp_hydrogens[i]), acceptor (tmp_acceptors[i]) matches the
            # distance (tmp_distances[i]).
            tmp_donors = []
            tmp_hydrogens = []
            tmp_acceptors = []
            tmp_distances = []
            for idx, distance in enumerate(distances):
                for h in donors[donors_idx[pairs[idx, 0]]]:
                    tmp_donors.append(donors_idx[pairs[idx, 0]])
                    tmp_hydrogens.append(h)
                    tmp_acceptors.append(acceptor[pairs[idx, 1]])
                    tmp_distances.append(distance)

        angles = np.rad2deg(
            calc_angles(
                self.u.atoms[tmp_donors].positions,
                self.u.atoms[tmp_hydrogens].positions,
                self.u.atoms[tmp_acceptors].positions,
                box=self.box,
            )
        )
        hbond_indices = np.where(angles > self.angle)[0]
        for index in hbond_indices:
            h = tmp_hydrogens[index]
            d = tmp_donors[index]
            a = tmp_acceptors[index]
            result.append(
                (
                    h,
                    d,
                    a,
                    self._expand_index(h),
                    self._expand_index(a),
                    tmp_distances[index],
                    angles[index],
                )
            )
        return result

    def _single_frame(self):
        self.timesteps.append(self._ts.time)
        self.box = self.u.dimensions if self.pbc else None

        if self.update_selection:
            self._update_selection()
        if len(self._s1) and len(self._s2):
            if self.update_water_selection:
                self._update_water_selection()
        else:
            self.results.network.append(defaultdict(dict))
            return

        selection_1 = []
        water_pool = defaultdict(list)
        next_round_water = set([])
        selection_2 = []

        if self.selection1_type in ("donor", "both"):
            # check for direct hbond from s1 to s2
            self.logger_debug("Selection 1 Donors <-> Selection 2 Acceptors")
            results = self._donor2acceptor(
                self._s1_donors_h, self._s1_h_donors, self._s2_acceptors
            )
            for line in results:
                (
                    h_index,
                    d_index,
                    a_index,
                    (h_resname, h_resid, h_name),
                    (a_resname, a_resid, a_name),
                    dist,
                    angle,
                ) = line
                water_pool[(a_resname, a_resid)] = None
                selection_1.append(
                    (h_index, d_index, a_index, None, dist, angle)
                )
                selection_2.append((a_resname, a_resid))
            if self.order > 0:
                self.logger_debug("Selection 1 Donors <-> Water Acceptors")
                results = self._donor2acceptor(
                    self._s1_donors_h, self._s1_h_donors, self._water_acceptors
                )
                for line in results:
                    (
                        h_index,
                        d_index,
                        a_index,
                        (h_resname, h_resid, h_name),
                        (a_resname, a_resid, a_name),
                        dist,
                        angle,
                    ) = line
                    selection_1.append(
                        (h_index, d_index, a_index, None, dist, angle)
                    )

                self.logger_debug("Water Donors <-> Selection 2 Acceptors")
                results = self._donor2acceptor(
                    self._water_donors_h,
                    self._water_h_donors,
                    self._s2_acceptors,
                )
                for line in results:
                    (
                        h_index,
                        d_index,
                        a_index,
                        (h_resname, h_resid, h_name),
                        (a_resname, a_resid, a_name),
                        dist,
                        angle,
                    ) = line
                    water_pool[(h_resname, h_resid)].append(
                        (h_index, d_index, a_index, None, dist, angle)
                    )
                    selection_2.append((a_resname, a_resid))

        if self.selection1_type in ("acceptor", "both"):
            self.logger_debug("Selection 2 Donors <-> Selection 1 Acceptors")
            results = self._donor2acceptor(
                self._s2_donors_h, self._s2_h_donors, self._s1_acceptors
            )
            for line in results:
                (
                    h_index,
                    d_index,
                    a_index,
                    (h_resname, h_resid, h_name),
                    (a_resname, a_resid, a_name),
                    dist,
                    angle,
                ) = line
                water_pool[(h_resname, h_resid)] = None
                selection_1.append(
                    (a_index, None, h_index, d_index, dist, angle)
                )
                selection_2.append((h_resname, h_resid))

            if self.order > 0:
                self.logger_debug("Selection 2 Donors <-> Water Acceptors")
                results = self._donor2acceptor(
                    self._s2_donors_h, self._s2_h_donors, self._water_acceptors
                )
                for line in results:
                    (
                        h_index,
                        d_index,
                        a_index,
                        (h_resname, h_resid, h_name),
                        (a_resname, a_resid, a_name),
                        dist,
                        angle,
                    ) = line
                    water_pool[(a_resname, a_resid)].append(
                        (a_index, None, h_index, d_index, dist, angle)
                    )
                    selection_2.append((h_resname, h_resid))

                self.logger_debug("Selection 1 Acceptors <-> Water Donors")
                results = self._donor2acceptor(
                    self._water_donors_h,
                    self._water_h_donors,
                    self._s1_acceptors,
                )
                for line in results:
                    (
                        h_index,
                        d_index,
                        a_index,
                        (h_resname, h_resid, h_name),
                        (a_resname, a_resid, a_name),
                        dist,
                        angle,
                    ) = line
                    selection_1.append(
                        (a_index, None, h_index, d_index, dist, angle)
                    )

        if self.order > 1:
            self.logger_debug("Water donor <-> Water Acceptors")
            results = self._donor2acceptor(
                self._water_donors_h,
                self._water_h_donors,
                self._water_acceptors,
            )
            for line in results:
                (
                    h_index,
                    d_index,
                    a_index,
                    (h_resname, h_resid, h_name),
                    (a_resname, a_resid, a_name),
                    dist,
                    angle,
                ) = line
                water_pool[(a_resname, a_resid)].append(
                    (a_index, None, h_index, d_index, dist, angle)
                )
                water_pool[(h_resname, h_resid)].append(
                    (h_index, d_index, a_index, None, dist, angle)
                )

        # solve the connectivity network
        # The following code attempt to generate a water network which is
        # formed by the class dict.
        # Suppose we have a water bridge connection ARG1 to ASP3 via the two
        # hydrogen bonds.
        #     [0,1,('ARG',1,'O'),  ('SOL',2,'HW1'),  3.0,180],
        #     [2,3,('SOL',2,'HW2'),('ASP',3,'OD1'),  3.0,180],
        # The resulting network will be
        # {(0,1,('ARG',1,'O'),  ('SOL',2,'HW1'),  3.0,180):
        # {(2,3,('SOL',2,'HW2'),('ASP',3,'OD1'),  3.0,180): None}}
        # Where the key of the a dict will be all the hydrogen bonds starting
        # from this nodes.
        # The corresponding value of a certain key will be a dictionary whose
        # key will be all the hydrogen bonds from
        # the destination of in the key.
        # If the value of a certain key is None, which means it is reaching
        # selection 2.

        result = {"start": defaultdict(dict), "water": defaultdict(dict)}

        def add_route(result, route):
            if len(route) == 1:
                result["start"][route[0]] = None
            else:
                # exclude the the selection which goes back to itself
                if sorted(route[0][0:3:2]) == sorted(route[-1][0:3:2]):
                    return

                # selection 2 to water
                result["water"][route[-1]] = None
                # water to water
                for i in range(1, len(route) - 1):
                    result["water"][route[i]][route[i + 1]] = result["water"][
                        route[i + 1]
                    ]
                # selection 1 to water
                result["start"][route[0]][route[1]] = result["water"][route[1]]

        def traverse_water_network(graph, node, end, route, maxdepth, result):
            if len(route) > self.order + 1:
                return
            else:
                if node in end:
                    # check if any duplication happens
                    heavy_atom = [line[3] or line[2] for line in route]
                    if len(heavy_atom) == len(set(heavy_atom)):
                        add_route(result, route)
                else:
                    for new_node in graph[node]:
                        new_route = route[:]
                        new_route.append(new_node)
                        new_node = self._expand_timeseries(
                            new_node, "sele1_sele2"
                        )[3][:2]
                        traverse_water_network(
                            graph, new_node, end, new_route, maxdepth, result
                        )

        for s1 in selection_1:
            route = [
                s1,
            ]
            next_mol = self._expand_timeseries(s1, "sele1_sele2")[3][:2]
            traverse_water_network(
                water_pool, next_mol, selection_2, route[:], self.order, result
            )

        self.results.network.append(result["start"])

    def _traverse_water_network(
        self,
        graph,
        current,
        analysis_func=None,
        output=None,
        link_func=None,
        **kwargs,
    ):
        """
        This function recursively traverses the water network
        self.results.network and finds the hydrogen bonds which connect the
        current atom to the next atom. The newly found hydrogen bond will be
        appended to the hydrogen bonds connecting the selection 1 to the
        current atom via link_func. When selection 2 is reached, the full list
        of hydrogen bonds connecting the selection 1 to selection 2 will be
        fed into analysis_func, which will then modify the output in place.

        :param graph: The connection network describes the connection between
        the atoms in the water network.
        :param current: The hydrogen bonds from selection 1 until now.
        :param analysis_func: The analysis function which is called to analysis
        the hydrogen bonds.
        :param output: where the result is stored.
        :param link_func: The new hydrogen bonds will be appended to current.
        :param kwargs: the keywords which are passed into the analysis_func.
        :return:
        """
        if link_func is None:
            # If no link_func is provided, the default link_func will be used
            link_func = self._full_link

        if graph is None:
            # if selection 2 is reached
            if analysis_func is not None:
                # the result is analysed by analysis_func which will change the
                # output
                analysis_func(current, output, self.u, **kwargs)
        else:
            # make sure no loop can occur
            if len(current) <= self.order:
                for node in graph:
                    # the new hydrogen bond will be added to the existing bonds
                    new = link_func(current, node)
                    self._traverse_water_network(
                        graph[node],
                        new,
                        analysis_func,
                        output,
                        link_func,
                        **kwargs,
                    )

    def _expand_index(self, index):
        """
        Expand the index into (resname, resid, name).
        """
        atom = self.u.atoms[index]
        return (atom.resname, atom.resid, atom.name)

    def _expand_timeseries(self, entry, output_format=None):
        """
        Expand the compact data format into the old timeseries form.
        The old is defined as the format for release up to 0.19.2.
        As is discussed in Issue #2177, the internal storage of the hydrogen
        bond information has been changed to the compact format.
        The function takes in the argument `output_format` to see which output
        format will be chosen.
        if `output_format` is not specified, the value will be taken from
        :attr:`output_format`.
        If `output_format` is 'sele1_sele2', the output will be the old water
        bridge analysis format::

          # donor from selection 1 to acceptor in selection 2
          [sele1_index, sele2_index,
           (sele1_resname, sele1_resid, sele1_name),
           (sele2_resname, sele2_resid, sele2_name), dist, angle]

        If `output_format` is 'donor_acceptor', the output will be the old
        hydrogen bond analysis format::

          # From donor to acceptor
          [donor_index, acceptor_index,
           (donor_resname, donor_resid, donor_name),
           (acceptor_resname, acceptor_resid, acceptor_name), dist, angle]
        """
        output_format = output_format or self.output_format
        # Expand the compact entry into atom1, which is the first index in the
        # output and atom2, which is the second
        # entry.
        atom1, heavy_atom1, atom2, heavy_atom2, dist, angle = entry
        if output_format == "sele1_sele2":
            # If the output format is the sele1_sele2, no change will be
            # executed
            atom1, atom2 = atom1, atom2
        elif output_format == "donor_acceptor":
            # If the output format is donor_acceptor, use heavy atom position
            # to check which is donor and which is
            # acceptor
            if heavy_atom1 is None:
                # atom1 is hydrogen bond acceptor and thus, the position of
                # atom1 and atom2 are swapped.
                atom1, atom2 = atom2, atom1
            else:
                # atom1 is hydrogen bond donor, position not swapped.
                atom1, atom2 = atom1, atom2
        else:
            raise KeyError(
                "Only 'sele1_sele2' or 'donor_acceptor' are allowed as output "
                "format"
            )

        return (
            atom1,
            atom2,
            self._expand_index(atom1),
            self._expand_index(atom2),
            dist,
            angle,
        )

    def _generate_timeseries(self, output_format=None):
        r"""Time series of water bridges.

        The output is generated per frame as is explained in
        :ref:`wb_Analysis_Timeseries`. The format of output can be changed via
        the output_format selection. If ``output_format="sele1_sele2"``, the
        hydrogen bond forms a directional link from selection 1 to selection 2.
        If ``output_format="donor_acceptor"``, for each hydrogen bond, the
        donor is always written before the acceptor.

        Note
        ----
        To find an acceptor atom in :attr:`Universe.atoms` by
        *index* one would use ``u.atoms[acceptor_index]``.

        .. versionchanged 0.20.0
           The :attr:`WaterBridgeAnalysis.timeseries` has been updated where
           the donor and acceptor string has been changed to tuple
           (resname string, resid, name_string).


        """
        output_format = output_format or self.output_format

        def analysis(current, output, *args, **kwargs):
            output = current

        timeseries = []
        for frame in self.results.network:
            new_frame = []
            self._traverse_water_network(
                frame,
                new_frame,
                analysis_func=analysis,
                output=new_frame,
                link_func=self._compact_link,
            )
            timeseries.append(
                [
                    self._expand_timeseries(entry, output_format)
                    for entry in new_frame
                ]
            )
        return timeseries

    def set_network(self, network):
        wmsg = (
            "The `set_network` method was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.network` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        self.results.network = network

    @classmethod
    def _full_link(self, output, node):
        """
        A function used in _traverse_water_network to add the new hydrogen
        bond to the existing bonds.
        :param output: The existing hydrogen bonds from selection 1
        :param node: The new hydrogen bond
        :return: The hydrogen bonds from selection 1 with the new hydrogen
        bond added
        """
        result = output[:]
        result.append(node)
        return result

    @classmethod
    def _compact_link(self, output, node):
        """
        A function used in _traverse_water_network to add the new hydrogen
        bond to the existing bonds. In this form no new list is created and
        thus, one bridge will only appear once.
        :param output: The existing hydrogen bonds from selection 1
        :param node: The new hydrogen bond
        :return: The hydrogen bonds from selection 1 with the new hydrogen
        bond added
        """
        output.append(node)
        return output

    def _count_by_type_analysis(self, current, output, *args, **kwargs):
        """
        Generates the key for count_by_type analysis.
        :return:
        """

        s1_index, to_index, s1, to_residue, dist, angle = (
            self._expand_timeseries(current[0])
        )
        s1_resname, s1_resid, s1_name = s1
        from_index, s2_index, from_residue, s2, dist, angle = (
            self._expand_timeseries(current[-1])
        )
        s2_resname, s2_resid, s2_name = s2
        key = (
            s1_index,
            s2_index,
            s1_resname,
            s1_resid,
            s1_name,
            s2_resname,
            s2_resid,
            s2_name,
        )
        output[key] += 1

    def count_by_type(self, analysis_func=None, **kwargs):
        """Counts the frequency of water bridge of a specific type.

        If one atom *A* from *selection 1* is linked to atom *B* from
        *selection 2* through one or more bridging waters, an entity will be
        created and the proportion of time that this linkage exists in the
        whole simulation will be calculated.

        The identification of a specific type of water bridge can be modified
        by supplying the analysis_func function. See :ref:`wb_count_by_type`
        for detail.

        Returns
        -------
        counts : list
            Returns a :class:`list` containing atom indices for *A* and
            *B*, residue names, residue numbers, atom names (for both A and B)
            and the fraction of the total time during which the water bridge
            was detected. This method returns None if method
            :meth:`WaterBridgeAnalysis.run` was not executed first.


        """
        output = None
        if analysis_func is None:
            analysis_func = self._count_by_type_analysis
            output = "combined"

        if self.results.network:
            length = len(self.results.network)
            result_dict = defaultdict(int)
            for frame in self.results.network:
                frame_dict = defaultdict(int)
                self._traverse_water_network(
                    frame,
                    [],
                    analysis_func=analysis_func,
                    output=frame_dict,
                    link_func=self._full_link,
                    **kwargs,
                )
                for key, value in frame_dict.items():
                    result_dict[key] += frame_dict[key]

            if output == "combined":
                result = [[i for i in key] for key in result_dict]
                [
                    result[i].append(result_dict[key] / length)
                    for i, key in enumerate(result_dict)
                ]
            else:
                result = [
                    (key, result_dict[key] / length) for key in result_dict
                ]
            return result
        else:
            return None

    def _count_by_time_analysis(self, current, output, *args, **kwargs):
        s1_index, to_index, s1, to_residue, dist, angle = (
            self._expand_timeseries(current[0])
        )
        s1_resname, s1_resid, s1_name = s1
        from_index, s2_index, from_residue, s2, dist, angle = (
            self._expand_timeseries(current[-1])
        )
        s2_resname, s2_resid, s2_name = s2
        key = (
            s1_index,
            s2_index,
            s1_resname,
            s1_resid,
            s1_name,
            s2_resname,
            s2_resid,
            s2_name,
        )
        output[key] += 1

    def count_by_time(self, analysis_func=None, **kwargs):
        """Counts the number of water bridges per timestep.

        The counting behaviour can be adjusted by supplying analysis_func.
        See :ref:`wb_count_by_time` for details.

        Returns
        -------
        counts : list
             Returns a time series ``N(t)`` where ``N`` is the total
             number of observed water bridges at time ``t``.

        """
        if analysis_func is None:
            analysis_func = self._count_by_time_analysis
        if self.results.network:
            result = []
            for time, frame in zip(self.timesteps, self.results.network):
                result_dict = defaultdict(int)
                self._traverse_water_network(
                    frame,
                    [],
                    analysis_func=analysis_func,
                    output=result_dict,
                    link_func=self._full_link,
                    **kwargs,
                )
                result.append(
                    (time, sum([result_dict[key] for key in result_dict]))
                )
            return result
        else:
            return None

    def _timesteps_by_type_analysis(self, current, output, *args, **kwargs):
        s1_index, to_index, s1, to_residue, dist, angle = (
            self._expand_timeseries(current[0])
        )
        s1_resname, s1_resid, s1_name = s1
        from_index, s2_index, from_residue, s2, dist, angle = (
            self._expand_timeseries(current[-1])
        )
        s2_resname, s2_resid, s2_name = s2
        key = (
            s1_index,
            s2_index,
            s1_resname,
            s1_resid,
            s1_name,
            s2_resname,
            s2_resid,
            s2_name,
        )
        output[key].append(kwargs.pop("time"))

    def timesteps_by_type(self, analysis_func=None, **kwargs):
        """Frames during which each water bridges existed, sorted by each water
        bridges.

        Processes :attr:`WaterBridgeAnalysis.results.network` and returns a
        :class:`list` containing atom indices, residue names, residue
        numbers (from selection 1 and selection 2) and each timestep at which
        the water bridge was detected.

        Similar to :meth:`~WaterBridgeAnalysis.count_by_type` and
        :meth:`~WaterBridgeAnalysis.count_by_time`, the behavior can be
        adjusted by supplying an analysis_func.

        Returns
        -------
        data : list

        """
        output = None
        if analysis_func is None:
            analysis_func = self._timesteps_by_type_analysis
            output = "combined"

        if self.results.network:
            result = defaultdict(list)
            if self.timesteps is None:
                timesteps = range(len(self.results.network))
            else:
                timesteps = self.timesteps
            for time, frame in zip(timesteps, self.results.network):
                self._traverse_water_network(
                    frame,
                    [],
                    analysis_func=analysis_func,
                    output=result,
                    link_func=self._full_link,
                    time=time,
                    **kwargs,
                )

            result_list = []
            for key, time_list in result.items():
                for time in time_list:
                    if output == "combined":
                        key = list(key)
                        key.append(time)
                        result_list.append(key)
                    else:
                        result_list.append((key, time))
            return result_list
        else:
            return None

    def generate_table(self, output_format=None):
        """Generate a normalised table of the results.

        Parameters
        ----------
        output_format : {'sele1_sele2', 'donor_acceptor'}
            The output format of the `table` can be changed a fashion similar
            to :attr:`WaterBridgeAnalysis.results.timeseries` by changing the
            labels of the columns of the participating atoms.

        Returns
        -------
        table : numpy.rec.recarray
            A "tidy" table with one hydrogen bond per row, labeled according to
            `output_format` and containing information of atom_1, atom_2,
            distance, and angle.

        .. versionchanged:: 2.0.0
           Return the generated table (as well as storing it as :attr:`table`).

        .. deprecated:: 2.0.0
           In release 3.0.0, :meth:`generate_table()` will _only_ return the
           table and no longer store it in :attr:`table`.
        """
        output_format = output_format or self.output_format
        if self.results.network == []:
            msg = "No data computed, do run() first."
            warnings.warn(msg, category=MissingDataWarning)
            logger.warning(msg)
            return None

        if (
            self.results.timeseries is not None
            and output_format == self.output_format
        ):
            timeseries = self.results.timeseries
        else:
            # Recompute timeseries with correct output format
            timeseries = self._generate_timeseries(output_format)

        num_records = np.sum([len(hframe) for hframe in timeseries])
        # build empty output table
        if output_format == "sele1_sele2":
            dtype = [
                ("time", float),
                ("sele1_index", int),
                ("sele2_index", int),
                ("sele1_resnm", "|U4"),
                ("sele1_resid", int),
                ("sele1_atom", "|U4"),
                ("sele2_resnm", "|U4"),
                ("sele2_resid", int),
                ("sele2_atom", "|U4"),
                ("distance", float),
                ("angle", float),
            ]
        elif output_format == "donor_acceptor":
            dtype = [
                ("time", float),
                ("donor_index", int),
                ("acceptor_index", int),
                ("donor_resnm", "|U4"),
                ("donor_resid", int),
                ("donor_atom", "|U4"),
                ("acceptor_resnm", "|U4"),
                ("acceptor_resid", int),
                ("acceptor_atom", "|U4"),
                ("distance", float),
                ("angle", float),
            ]

        # according to Lukas' notes below, using a recarray at this stage is
        # ineffective and speedups of ~x10 can be achieved by filling a
        # standard array, like this:
        out = np.empty((num_records,), dtype=dtype)
        cursor = 0  # current row
        for t, hframe in zip(self.timesteps, timeseries):
            for (
                donor_index,
                acceptor_index,
                donor,
                acceptor,
                distance,
                angle,
            ) in hframe:
                # donor|acceptor = (resname, resid, atomid)
                out[cursor] = (
                    (t, donor_index, acceptor_index)
                    + donor
                    + acceptor
                    + (distance, angle)
                )
                cursor += 1
        assert (
            cursor == num_records
        ), "Internal Error: Not all wb records stored"
        table = out.view(np.rec.recarray)
        logger.debug(
            "WBridge: Stored results as table with %(num_records)d entries.",
            vars(),
        )
        self.table = table

        return table

    def _conclude(self):
        self.results.timeseries = self._generate_timeseries()

    @property
    def network(self):
        wmsg = (
            "The `network` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.network` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.network

    @property
    def timeseries(self):
        wmsg = (
            "The `timeseries` attribute was deprecated in MDAnalysis "
            "2.0.0 and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.timeseries` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.timeseries



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""AMBER trajectories --- :mod:`MDAnalysis.coordinates.TRJ`
========================================================

AMBER_ can write :ref:`ASCII trajectories<ascii-trajectories>` ("traj") and
:ref:`binary trajectories<netcdf-trajectories>` ("netcdf"). MDAnalysis supports
reading of both formats and writing for the binary trajectories.

Note
----
Support for AMBER is still somewhat *experimental* and feedback and
contributions are highly appreciated. Use the `Issue Tracker`_ or get in touch
on the `GitHub Discussions`_.


.. rubric:: Units

AMBER trajectories are assumed to be in the following units:

* lengths in Angstrom (Å)
* time in ps (but see below)


.. _netcdf-trajectories:

Binary NetCDF trajectories
--------------------------

The `AMBER netcdf`_ format make use of NetCDF_ (Network Common Data
Form) format. Such binary trajectories are recognized in MDAnalysis by
the '.ncdf' suffix and read by the :class:`NCDFReader`.

Binary trajectories can also contain velocities and forces, and can record the
exact time
step. In principle, the trajectories can be in different units than the AMBER
defaults of ångström and picoseconds but at the moment MDAnalysis only supports
those and will raise a :exc:`NotImplementedError` if anything else is detected.

.. autoclass:: NCDFReader
   :members:

.. autoclass:: NCDFWriter
   :members:

.. autoclass:: NCDFPicklable
   :members:

.. _ascii-trajectories:

ASCII TRAJ trajectories
-----------------------

ASCII AMBER_ TRJ coordinate files (as defined in `AMBER TRJ format`_)
are handled by the :class:`TRJReader`. It is also possible to directly
read *bzip2* or *gzip* compressed files.

AMBER ASCII trajectories are recognised by the suffix '.trj',
'.mdcrd' or '.crdbox (possibly with an additional '.gz' or '.bz2').

.. Note::

   In the AMBER community, these trajectories are often saved with the
   suffix '.crd' but this extension conflicts with the CHARMM CRD
   format and MDAnalysis *will not correctly autodetect AMBER ".crd"
   trajectories*. Instead, explicitly provide the ``format="TRJ"``
   argument to :class:`~MDAnalysis.core.universe.Universe`::

     u = MDAnalysis.Universe("top.prmtop", "traj.crd", format="TRJ")

   In this way, the AMBER :class:`TRJReader` is used.


.. rubric:: Limitations

* Periodic boxes are only stored as box lengths A, B, C in an AMBER
  trajectory; the reader always assumes that these are orthorhombic
  boxes.

* The trajectory does not contain time information so we simply set
  the time step to 1 ps (or the user could provide it as kwarg *dt*)

* Trajectories with fewer than 4 atoms probably fail to be read (BUG).

* If the trajectory contains exactly *one* atom then it is always
  assumed to be non-periodic (for technical reasons).

* Velocities are currently *not supported* as ASCII trajectories.

.. autoclass:: TRJReader
   :members:



.. Links

.. _AMBER: http://ambermd.org
.. _AMBER TRJ format: https://ambermd.org/FileFormats.php#trajectory
..    The formats page was archived as
..    http://www.webcitation.org/query?url=http%3A%2F%2Fambermd.org%2Fformats.html&date=2018-02-11
..    Use the archived version if the original (https://ambermd.org/FileFormats.php#trajectory) disappears. [orbeckst]
.. _AMBER netcdf format: http://ambermd.org/netcdf/nctraj.xhtml
..    The formats page was archived as
..    http://www.webcitation.org/query?url=http%3A%2F%2Fambermd.org%2Fnetcdf%2Fnctraj.xhtml&date=2018-02-11
..    Use the archived version if the original disappears. [orbeckst]
.. _AMBER netcdf: http://ambermd.org/netcdf/nctraj.xhtml
.. _NetCDF: http://www.unidata.ucar.edu/software/netcdf
.. _Issue Tracker: https://github.com/MDAnalysis/mdanalysis/issues
.. _GitHub Discussions: https://github.com/MDAnalysis/mdanalysis/discussions

"""
import scipy.io.netcdf
import numpy as np
import warnings
import errno
import logging
from math import isclose

import MDAnalysis
from .timestep import Timestep
from . import base
from ..lib import util
from ..lib.util import store_init_arguments
logger = logging.getLogger("MDAnalysis.coordinates.AMBER")


try:
    import netCDF4
except ImportError:
    netCDF4 = None
    logger.warning("netCDF4 is not available. Writing AMBER ncdf files will be slow.")


class TRJReader(base.ReaderBase):
    """AMBER trajectory reader.

    Reads the ASCII formatted `AMBER TRJ format`_. Periodic box information
    is auto-detected.

    The number of atoms in a timestep *must* be provided in the `n_atoms`
    keyword because it is not stored in the trajectory header and cannot be
    reliably autodetected. The constructor raises a :exc:`ValueError` if
    `n_atoms` is left at its default value of ``None``.

    The length of a timestep is not stored in the trajectory itself but can
    be set by passing the `dt` keyword argument to the constructor; it
    is assumed to be in ps. The default value is 1 ps.

    .. _AMBER TRJ format: https://ambermd.org/FileFormats.php#trajectory

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based.
       kwarg `delta` renamed to `dt`, for uniformity with other Readers
    """
    format = ['TRJ', 'MDCRD', 'CRDBOX']
    units = {'time': 'ps', 'length': 'Angstrom'}
    _Timestep = Timestep

    @store_init_arguments
    def __init__(self, filename, n_atoms=None, **kwargs):
        super(TRJReader, self).__init__(filename, **kwargs)
        if n_atoms is None:
            raise ValueError("AMBER TRJ reader REQUIRES the n_atoms keyword")
        self._n_atoms = n_atoms
        self._n_frames = None

        self.trjfile = None  # have _read_next_timestep() open it properly!
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        # FORMAT(10F8.3)  (X(i), Y(i), Z(i), i=1,NATOM)
        self.default_line_parser = util.FORTRANReader("10F8.3")
        self.lines_per_frame = int(np.ceil(3.0 * self.n_atoms / len(
            self.default_line_parser)))
        # The last line per frame might have fewer than 10
        # We determine right away what parser we need for the last
        # line because it will be the same for all frames.
        last_per_line = 3 * self.n_atoms % len(self.default_line_parser)
        self.last_line_parser = util.FORTRANReader("{0:d}F8.3".format(
            last_per_line))

        # FORMAT(10F8.3)  BOX(1), BOX(2), BOX(3)
        # is this always on a separate line??
        self.box_line_parser = util.FORTRANReader("3F8.3")

        # Now check for box
        self._detect_amber_box()

        # open file, read first frame
        self._read_next_timestep()

    def _read_frame(self, frame):
        if self.trjfile is None:
            self.open_trajectory()
        self.trjfile.seek(self._offsets[frame])
        self.ts.frame = frame - 1  # gets +1'd in _read_next
        return self._read_next_timestep()

    def _read_next_timestep(self):
        # FORMAT(10F8.3)  (X(i), Y(i), Z(i), i=1,NATOM)
        ts = self.ts
        if self.trjfile is None:
            self.open_trajectory()

        # Read coordinat frame:
        # coordinates = numpy.zeros(3*self.n_atoms, dtype=np.float32)
        _coords = []
        for number, line in enumerate(self.trjfile):
            try:
                _coords.extend(self.default_line_parser.read(line))
            except ValueError:
                # less than 10 entries on the line:
                _coords.extend(self.last_line_parser.read(line))
            if number == self.lines_per_frame - 1:
                # read all atoms that are there in this frame
                break
        if _coords == []:
            # at the end of the stream (the loop has not been entered)
            raise EOFError

        # Read box information
        if self.periodic:
            line = next(self.trjfile)
            box = self.box_line_parser.read(line)
            ts.dimensions = box + [90., 90., 90.]  # assumed

        # probably slow ... could be optimized by storing the coordinates in
        # X,Y,Z lists or directly filling the array; the array/reshape is not
        # good because it creates an intermediate array
        ts._pos[:] = np.array(_coords).reshape(self.n_atoms, 3)
        ts.frame += 1
        return ts

    def _detect_amber_box(self):
        """Detecting a box in a AMBER trajectory

        Rewind trajectory and check for potential box data
        after the first frame.

        Set :attr:`TRJReader.periodic` to ``True`` if box was
        found, ``False`` otherwise.

        Only run at the beginning as it *rewinds* the trajctory.

         - see if there's data after the atoms have been read that looks
           like::

             FORMAT(10F8.3)  BOX(1), BOX(2), BOX(3)
             BOX    : size of periodic box

         - this WILL fail if we have exactly 1 atom in the trajectory because
           there's no way to distinguish the coordinates from the box
           so for 1 atom we always assume no box

        .. TODO:: needs a Timestep that knows about AMBER unitcells!
        """
        if self.n_atoms == 1:
            # for 1 atom we cannot detect the box with the current approach
            self.periodic = False  # see _read_next_timestep()!
            wmsg = "Trajectory contains a single atom: assuming periodic=False"
            warnings.warn(wmsg)
            return False

        self._reopen()
        self.periodic = False  # make sure that only coordinates are read
        self._read_next_timestep()
        ts = self.ts
        # TODO: what do we do with 1-frame trajectories? Try..except EOFError?
        line = next(self.trjfile)
        nentries = self.default_line_parser.number_of_matches(line)
        if nentries == 3:
            self.periodic = True
            ts.dimensions = self.box_line_parser.read(line) + [90., 90., 90.]
        else:
            self.periodic = False
        self.close()
        return self.periodic

    @property
    def n_frames(self):
        """Number of frames (obtained from reading the whole trajectory)."""
        if self._n_frames is not None:  # return cached value
            return self._n_frames
        try:
            self._n_frames = self._read_trj_n_frames(self.filename)
        except IOError:
            return 0
        else:
            return self._n_frames

    def _read_trj_n_frames(self, filename):
        lpf = self.lines_per_frame
        if self.periodic:
            lpf += 1

        self._offsets = offsets = []
        counter = 0
        with util.openany(self.filename) as f:
            line = f.readline()  # ignore first line
            while line:
                if counter % lpf == 0:
                    offsets.append(f.tell())
                line = f.readline()
                counter += 1
        offsets.pop()  # last offset is EOF
        return len(offsets)

    @property
    def n_atoms(self):
        return self._n_atoms

    def _reopen(self):
        self.close()
        self.open_trajectory()

    def open_trajectory(self):
        """Open the trajectory for reading and load first frame."""
        self.trjfile = util.anyopen(self.filename)
        self.header = self.trjfile.readline()  # ignore first line
        if len(self.header.rstrip()) > 80:
            # Chimera uses this check
            raise OSError(
                "Header of AMBER formatted trajectory has more than 80 chars. "
                "This is probably not a AMBER trajectory.")
        # reset ts
        ts = self.ts
        ts.frame = -1

        return self.trjfile

    def close(self):
        """Close trj trajectory file if it was open."""
        if self.trjfile is None:
            return
        self.trjfile.close()
        self.trjfile = None


class NCDFReader(base.ReaderBase):
    """Reader for `AMBER NETCDF format`_ (version 1.0).

    AMBER binary trajectories are automatically recognised by the
    file extension ".ncdf".

    The number of atoms (`n_atoms`) does not have to be provided as it can
    be read from the trajectory. The trajectory reader can randomly access
    frames and therefore supports direct indexing (with 0-based frame
    indices) and full-feature trajectory iteration, including slicing.

    Velocities are autodetected and read into the
    :attr:`Timestep._velocities` attribute.

    Forces are autodetected and read into the
    :attr:`Timestep._forces` attribute.

    Periodic unit cell information is detected and used to populate the
    :attr:`Timestep.dimensions` attribute. (If no unit cell is available in
    the trajectory, then :attr:`Timestep.dimensions` will return
    ``[0,0,0,0,0,0]``).

    Current limitations:

    * only trajectories with time in ps and lengths in Angstroem are processed

    The NCDF reader uses :mod:`scipy.io.netcdf` and therefore :mod:`scipy` must
    be installed. It supports the *mmap* keyword argument (when reading):
    ``mmap=True`` is memory efficient and directly maps the trajectory on disk
    to memory (using the :class:`~mmap.mmap`); ``mmap=False`` may consume large
    amounts of memory because it loads the whole trajectory into memory but it
    might be faster. The default is ``mmap=None`` and then default behavior of
    :class:`scipy.io.netcdf_file` prevails, i.e. ``True`` when
    *filename* is a file name, ``False`` when *filename* is a file-like object.

    .. _AMBER NETCDF format: http://ambermd.org/netcdf/nctraj.xhtml

    See Also
    --------
    :class:`NCDFWriter`


    .. versionadded: 0.7.6
    .. versionchanged:: 0.10.0
       Added ability to read Forces
    .. versionchanged:: 0.11.0
       Frame labels now 0-based instead of 1-based.
       kwarg `delta` renamed to `dt`, for uniformity with other Readers.
    .. versionchanged:: 0.17.0
       Uses :mod:`scipy.io.netcdf` and supports the *mmap* kwarg.
    .. versionchanged:: 0.20.0
       Now reads scale_factors for all expected AMBER convention variables.
       Timestep variables now adhere standard MDAnalysis units, with lengths
       of angstrom, time of ps, velocity of angstrom/ps and force of
       kJ/(mol*Angstrom). It is noted that with 0.19.2 and earlier versions,
       velocities would have often been reported in values of angstrom/AKMA
       time units instead (Issue #2323).
    .. versionchanged:: 1.0.0
       Support for reading `degrees` units for `cell_angles` has now been
       removed (Issue #2327)
    .. versionchanged:: 2.0.0
       Now use a picklable :class:`scipy.io.netcdf_file`--
       :class:`NCDFPicklable`.
       Reading of `dt` now defaults to 1.0 ps if `dt` cannot be extracted from
       the first two frames of the trajectory.
       :meth:`Writer` now also sets `convert_units`, `velocities`, `forces` and
       `scale_factor` information for the :class:`NCDFWriter`.

    """

    format = ['NCDF', 'NC']
    multiframe = True
    version = "1.0"
    units = {'time': 'ps',
             'length': 'Angstrom',
             'velocity': 'Angstrom/ps',
             'force': 'kcal/(mol*Angstrom)'}

    _Timestep = Timestep

    @store_init_arguments
    def __init__(self, filename, n_atoms=None, mmap=None, **kwargs):

        self._mmap = mmap

        super(NCDFReader, self).__init__(filename, **kwargs)

        # ensure maskandscale is off so we don't end up double scaling
        self.trjfile = NCDFPicklable(self.filename,
                                     mmap=self._mmap,
                                     maskandscale=False)

        # AMBER NetCDF files should always have a convention
        try:
            conventions = self.trjfile.Conventions
            if not ('AMBER' in conventions.decode('utf-8').split(',') or
                    'AMBER' in conventions.decode('utf-8').split()):
                errmsg = ("NCDF trajectory {0} does not conform to AMBER "
                          "specifications, "
                          "http://ambermd.org/netcdf/nctraj.xhtml "
                          "('AMBER' must be one of the token in attribute "
                          "Conventions)".format(self.filename))
                logger.fatal(errmsg)
                raise TypeError(errmsg)
        except AttributeError:
            errmsg = "NCDF trajectory {0} is missing Conventions".format(
                      self.filename)
            logger.fatal(errmsg)
            raise ValueError(errmsg) from None

        # AMBER NetCDF files should also have a ConventionVersion
        try:
            ConventionVersion = self.trjfile.ConventionVersion.decode('utf-8')
            if not ConventionVersion == self.version:
                wmsg = ("NCDF trajectory format is {0!s} but the reader "
                        "implements format {1!s}".format(
                         ConventionVersion, self.version))
                warnings.warn(wmsg)
                logger.warning(wmsg)
        except AttributeError:
            errmsg = "NCDF trajectory {0} is missing ConventionVersion".format(
                      self.filename)
            raise ValueError(errmsg) from None

        # The AMBER NetCDF standard enforces 64 bit offsets
        if not self.trjfile.version_byte == 2:
            errmsg = ("NCDF trajectory {0} does not conform to AMBER "
                      "specifications, as detailed in "
                      "https://ambermd.org/netcdf/nctraj.xhtml "
                      "(NetCDF file does not use 64 bit offsets "
                      "[version_byte = 2])".format(self.filename))
            logger.fatal(errmsg)
            raise TypeError(errmsg)

        # The AMBER NetCDF standard enforces 3D coordinates
        try:
            if not self.trjfile.dimensions['spatial'] == 3:
                errmsg = "Incorrect spatial value for NCDF trajectory file"
                raise TypeError(errmsg)
        except KeyError:
            errmsg = "NCDF trajectory does not contain spatial dimension"
            raise ValueError(errmsg) from None

        # AMBER NetCDF specs require program and programVersion. Warn users
        # if those attributes do not exist
        if not (hasattr(self.trjfile, 'program') and
                hasattr(self.trjfile, 'programVersion')):
            wmsg = ("NCDF trajectory {0} may not fully adhere to AMBER "
                    "standards as either the `program` or `programVersion` "
                    "attributes are missing".format(self.filename))
            warnings.warn(wmsg)
            logger.warning(wmsg)

        try:
            self.n_atoms = self.trjfile.dimensions['atom']
            if n_atoms is not None and n_atoms != self.n_atoms:
                errmsg = ("Supplied n_atoms ({0}) != natom from ncdf ({1}). "
                          "Note: n_atoms can be None and then the ncdf value "
                          "is used!".format(n_atoms, self.n_atoms))
                raise ValueError(errmsg)
        except KeyError:
            errmsg = ("NCDF trajectory {0} does not contain atom "
                      "information".format(self.filename))
            raise ValueError(errmsg) from None

        try:
            self.n_frames = self.trjfile.dimensions['frame']
            # example trajectory when read with scipy.io.netcdf has
            # dimensions['frame'] == None (indicating a record dimension that
            # can grow) whereas if read with netCDF4 I get
            # len(dimensions['frame']) ==  10: in any case, we need to get
            # the number of frames from somewhere such as the time variable:
            if self.n_frames is None:
                self.n_frames = self.trjfile.variables['coordinates'].shape[0]
        except KeyError:
            errmsg = (f"NCDF trajectory {self.filename} does not contain "
                      f"frame information")
            raise ValueError(errmsg) from None

        try:
            self.remarks = self.trjfile.title
        except AttributeError:
            self.remarks = ""
        # other metadata (*= requd):
        # - application           AMBER
        #

        # checks for not-implemented features (other units would need to be
        # hacked into MDAnalysis.units)
        try:
            self._verify_units(self.trjfile.variables['time'].units, 'picosecond')
            self.has_time = True
        except KeyError:
            self.has_time = False
            wmsg = ("NCDF trajectory does not contain `time` information;"
                    " `time` will be set as an increasing index")  
            warnings.warn(wmsg)
            logger.warning(wmsg)


        self._verify_units(self.trjfile.variables['coordinates'].units,
                           'angstrom')

        # Check for scale_factor attributes for all data variables and
        # store this to multiply through later (Issue #2323)
        self.scale_factors = {'time': None,
                              'cell_lengths': None,
                              'cell_angles': None,
                              'coordinates': None,
                              'velocities': None,
                              'forces': None}

        for variable in self.trjfile.variables:
            if hasattr(self.trjfile.variables[variable], 'scale_factor'):
                if variable in self.scale_factors:
                    scale_factor = self.trjfile.variables[variable].scale_factor
                    if not isinstance(scale_factor, (float, np.floating)):
                        raise TypeError(f"{scale_factor} is not a float")
                    self.scale_factors[variable] = scale_factor
                else:
                    errmsg = ("scale_factors for variable {0} are "
                              "not implemented".format(variable))
                    raise NotImplementedError(errmsg)

        self.has_velocities = 'velocities' in self.trjfile.variables
        if self.has_velocities:
            self._verify_units(self.trjfile.variables['velocities'].units,
                               'angstrom/picosecond')

        self.has_forces = 'forces' in self.trjfile.variables
        if self.has_forces:
            self._verify_units(self.trjfile.variables['forces'].units,
                               'kilocalorie/mole/angstrom')

        self.periodic = 'cell_lengths' in self.trjfile.variables
        if self.periodic:
            self._verify_units(self.trjfile.variables['cell_lengths'].units,
                               'angstrom')
            # As of v1.0.0 only `degree` is accepted as a unit
            cell_angle_units = self.trjfile.variables['cell_angles'].units
            self._verify_units(cell_angle_units, 'degree')

        self._current_frame = 0

        self.ts = self._Timestep(self.n_atoms,
                                 velocities=self.has_velocities,
                                 forces=self.has_forces,
                                 reader=self,  # for dt
                                 **self._ts_kwargs)

        # load first data frame
        self._read_frame(0)

    @staticmethod
    def _verify_units(eval_unit, expected_units):
        if eval_unit.decode('utf-8') != expected_units:
            errmsg = ("NETCDFReader currently assumes that the trajectory "
                      "was written in units of {0} instead of {1}".format(
                       eval_unit.decode('utf-8'), expected_units))
            raise NotImplementedError(errmsg)

    @staticmethod
    def parse_n_atoms(filename, **kwargs):
        with scipy.io.netcdf_file(filename, mmap=None) as f:
            n_atoms = f.dimensions['atom']
        return n_atoms

    def _get_var_and_scale(self, variable, frame):
        """Helper function to get variable at given frame from NETCDF file and
        scale if necessary.

        Note
        ----
        If scale_factor is 1.0 within numerical precision then we don't apply
        the scaling.
        """
        scale_factor = self.scale_factors[variable]
        if scale_factor is None or isclose(scale_factor, 1):
            return self.trjfile.variables[variable][frame]
        else:
            return self.trjfile.variables[variable][frame] * scale_factor

    def _read_frame(self, frame):
        ts = self.ts

        if self.trjfile is None:
            raise IOError("Trajectory is closed")
        if np.dtype(type(frame)) != np.dtype(int):
            # convention... for netcdf could also be a slice
            raise TypeError("frame must be a positive integer or zero")
        if frame >= self.n_frames or frame < 0:
            raise IndexError("frame index must be 0 <= frame < {0}".format(
                self.n_frames))
        # note: self.trjfile.variables['coordinates'].shape == (frames, n_atoms, 3)
        ts._pos[:] = self._get_var_and_scale('coordinates', frame)
        if self.has_time:
            ts.time = self._get_var_and_scale('time', frame)
        if self.has_velocities:
            ts._velocities[:] = self._get_var_and_scale('velocities', frame)
        if self.has_forces:
            ts._forces[:] = self._get_var_and_scale('forces', frame)
        if self.periodic:
            unitcell = np.zeros(6)
            unitcell[:3] = self._get_var_and_scale('cell_lengths', frame)
            unitcell[3:] = self._get_var_and_scale('cell_angles', frame)
            ts.dimensions = unitcell
        if self.convert_units:
            self.convert_pos_from_native(ts._pos)  # in-place !
            self.convert_time_from_native(
                ts.time)  # in-place ! (hope this works...)
            if self.has_velocities:
                self.convert_velocities_from_native(ts._velocities,
                                                    inplace=True)
            if self.has_forces:
                self.convert_forces_from_native(ts._forces, inplace=True)
            if self.periodic:
                # in-place ! (only lengths)
                self.convert_pos_from_native(ts.dimensions[:3])
        ts.frame = frame  # frame labels are 0-based
        self._current_frame = frame
        return ts

    def _reopen(self):
        self._current_frame = -1

    def _read_next_timestep(self, ts=None):
        if ts is None:
            ts = self.ts
        try:
            return self._read_frame(self._current_frame + 1)
        except IndexError:
            raise IOError from None

    def _get_dt(self):
        """Gets dt based on the time difference between the first and second
        frame. If missing (i.e. an IndexError is triggered), raises an
        AttributeError which triggers the default 1 ps return of dt().
        """
        try:
            t1 = self.trjfile.variables['time'][1]
            t0 = self.trjfile.variables['time'][0]
        except (IndexError, KeyError):
            raise AttributeError
        return t1 - t0

    def close(self):
        """Close trajectory; any further access will raise an :exc:`IOError`.

        .. Note:: The underlying :mod:`scipy.io.netcdf` module may open netcdf
                  files with :class:`~mmap.mmap` if ``mmap=True`` was
                  set. Hence *any* reference to an array *must* be removed
                  before the file can be closed.

        """
        if self.trjfile is not None:
            self.trjfile.close()
            self.trjfile = None

    def Writer(self, filename, **kwargs):
        """Returns a NCDFWriter for `filename` with the same parameters as this NCDF.

        All values can be changed through keyword arguments.

        Parameters
        ----------
        filename : str
            filename of the output NCDF trajectory
        n_atoms : int (optional)
            number of atoms
        remarks : str (optional)
            string that is stored in the title field
        convert_units : bool (optional)
            ``True``: units are converted to the AMBER base format
        velocities : bool (optional)
            Write velocities into the trajectory
        forces : bool (optional)
            Write forces into the trajectory
        scale_time : float (optional)
            Scale factor for time units
        scale_cell_lengths : float (optional)
            Scale factor for cell lengths
        scale_cell_angles : float (optional)
            Scale factor for cell angles
        scale_coordinates : float (optional)
            Scale factor for coordinates
        scale_velocities : float (optional)
            Scale factor for velocities
        scale_forces : float (optional)
            Scale factor for forces

        Returns
        -------
        :class:`NCDFWriter`
        """
        n_atoms = kwargs.pop('n_atoms', self.n_atoms)
        kwargs.setdefault('remarks', self.remarks)
        kwargs.setdefault('convert_units', self.convert_units)
        kwargs.setdefault('velocities', self.has_velocities)
        kwargs.setdefault('forces', self.has_forces)
        for key in self.scale_factors:
            kwargs.setdefault(f'scale_{key}', self.scale_factors[key])
        return NCDFWriter(filename, n_atoms, **kwargs)


class NCDFWriter(base.WriterBase):
    """Writer for `AMBER NETCDF format`_ (version 1.0).

    AMBER binary trajectories are automatically recognised by the
    file extension ".ncdf" or ".nc".

    Velocities are written out if they are detected in the input
    :class:`Timestep`. The trajectories are always written with ångström
    for the lengths and picoseconds for the time (and hence Å/ps for
    velocities and kilocalorie/mole/Å for forces).

    Scale factor variables for time, velocities, cell lengths, cell angles,
    coordinates, velocities, or forces can be passed into the writer. If so,
    they will be written to the NetCDF file. In this case, the trajectory data
    will be written to the NetCDF file divided by the scale factor value. By
    default, scale factor variables are not written. The only exception is for
    velocities, where it is set to 20.455, replicating the default behaviour of
    AMBER.

    Unit cell information is written if available.

    .. _AMBER NETCDF format: http://ambermd.org/netcdf/nctraj.xhtml


    Parameters
    ----------
    filename : str
        name of output file
    n_atoms : int
        number of atoms in trajectory file
    convert_units : bool (optional)
        ``True``: units are converted to the AMBER base format; [``True``]
    velocities : bool (optional)
        Write velocities into the trajectory [``False``]
    forces : bool (optional)
        Write forces into the trajectory [``False``]
    scale_time : float (optional)
        Scale factor for time units [`None`]
    scale_cell_lengths : float (optional)
        Scale factor for cell lengths [``None``]
    scale_cell_angles : float (optional)
        Scale factor for cell angles [``None``]
    scale_coordinates : float (optional)
        Scale factor for coordinates [``None``]
    scale_velocities : float (optional)
        Scale factor for velocities [20.455]
    scale_forces : float (optional)
        Scale factor for forces [``None``]


    Note
    ----
    MDAnalysis uses :mod:`scipy.io.netcdf` to access AMBER files, which are in
    netcdf 3 format. Although :mod:`scipy.io.netcdf` is very fast at reading
    these files, it is *very* slow when writing, and it becomes slower the
    longer the files are. On the other hand, the netCDF4_ package (which
    requires the compiled netcdf library to be installed) is fast at writing
    but slow at reading. Therefore, we try to use :mod:`netCDF4` for writing if
    available but otherwise fall back to the slower :mod:`scipy.io.netcdf`.

    **AMBER users** might have a hard time getting netCDF4 to work with a
    conda-based installation (as discussed in `Issue #506`_) because of the way
    that AMBER itself handles netcdf: When the AMBER environment is loaded, the
    following can happen when trying to import netCDF4::

      >>> import netCDF4
      Traceback (most recent call last):
        File "<string>", line 1, in <module>
        File "/scratch2/miniconda/envs/py35/lib/python3.5/site-packages/netCDF4/__init__.py", line 3, in <module>
          from ._netCDF4 import *
      ImportError: /scratch2/miniconda/envs/py35/lib/python3.5/site-packages/netCDF4/_netCDF4.cpython-35m-x86_64-linux-gnu.so: undefined symbol: nc_inq_var_fletcher32

    The reason for this (figured out via :program:`ldd`) is that AMBER builds
    its own NetCDF library that it now inserts into :envvar:`LD_LIBRARY_PATH`
    *without the NetCDF4 API and HDF5 bindings*. Since the conda version of
    :mod:`netCDF4` was built against the full NetCDF package, the one
    :program:`ld` tries to link to at runtime (because AMBER requires
    :envvar:`LD_LIBRARY_PATH`) is missing some symbols. Removing AMBER from the
    environment fixes the import but is not really a convenient solution for
    users of AMBER.

    At the moment there is no obvious solution if one wants to use
    :mod:`netCDF4` and AMBER in the same shell session. If you need the fast
    writing capabilities of :mod:`netCDF4` then you need to unload your AMBER
    environment before importing MDAnalysis.


    .. _netCDF4: https://unidata.github.io/netcdf4-python/
    .. _`Issue #506`:
       https://github.com/MDAnalysis/mdanalysis/issues/506#issuecomment-225081416

    Raises
    ------
    FutureWarning
        When writing. The :class:`NCDFWriter` currently does not write any
        `scale_factor` values for the data variables. Whilst not in breach
        of the AMBER NetCDF standard, this behaviour differs from that of
        most AMBER writers, especially for velocities which usually have a
        `scale_factor` of 20.455. In MDAnalysis 2.0, the :class:`NCDFWriter`
        will enforce `scale_factor` writing to either match user inputs (either
        manually defined or via the :class:`NCDFReader`) or those as written by
        AmberTools's :program:`sander` under default operation.

    See Also
    --------
    :class:`NCDFReader`


    .. versionadded: 0.7.6
    .. versionchanged:: 0.10.0
       Added ability to write velocities and forces
    .. versionchanged:: 0.11.0
       kwarg `delta` renamed to `dt`, for uniformity with other Readers
    .. versionchanged:: 0.17.0
       Use fast :mod:`netCDF4` for writing but fall back to slow
       :mod:`scipy.io.netcdf` if :mod:`netCDF4` is not available.
    .. versionchanged:: 0.20.1
       Changes the `cell_angles` unit to the AMBER NetCDF convention standard
       of `degree` instead of the `degrees` written in previous version of
       MDAnalysis (Issue #2327).
    .. versionchanged:: 2.0.0
       ``dt``, ``start``, and ``step`` keywords were unused and are no longer
       set.
       Writing of ``scale_factor`` values has now been implemented. By default
       only velocities write a scale_factor of 20.455 (echoing the behaviour
       seen from AMBER).

    """

    format = ['NC', 'NCDF']
    multiframe = True
    version = "1.0"
    units = {'time': 'ps',
             'length': 'Angstrom',
             'velocity': 'Angstrom/ps',
             'force': 'kcal/(mol*Angstrom)'}

    def __init__(self, filename, n_atoms, remarks=None, convert_units=True,
                 velocities=False, forces=False, scale_time=None,
                 scale_cell_lengths=None, scale_cell_angles=None,
                 scale_coordinates=None, scale_velocities=None,
                 scale_forces=None, **kwargs):
        self.filename = filename
        if n_atoms == 0:
            raise ValueError("NCDFWriter: no atoms in output trajectory")
        self.n_atoms = n_atoms
        # convert length and time to base units on the fly?
        self.convert_units = convert_units

        self.remarks = remarks or "AMBER NetCDF format (MDAnalysis.coordinates.trj.NCDFWriter)"

        self._first_frame = True  # signals to open trajectory
        self.trjfile = None  # open on first write with _init_netcdf()
        self.periodic = None  # detect on first write
        self.has_velocities = velocities
        self.has_forces = forces

        self.scale_factors = {
                'time': scale_time,
                'cell_lengths': scale_cell_lengths,
                'cell_angles': scale_cell_angles,
                'coordinates': scale_coordinates,
                'velocities': scale_velocities,
                'forces': scale_forces}
        # NCDF standard enforces float scale_factors
        for value in self.scale_factors.values():
            if (value is not None) and (
                    not isinstance(value, (float, np.floating))):
                errmsg = f"scale_factor {value} is not a float"
                raise TypeError(errmsg)

        self.curr_frame = 0

    def _init_netcdf(self, periodic=True):
        """Initialize netcdf AMBER 1.0 trajectory.

        The trajectory is opened when the first frame is written
        because that is the earliest time that we can detect if the
        output should contain periodicity information (i.e. the unit
        cell dimensions).

        Based on Joshua Adelman's `netcdf4storage.py`_ in `Issue 109`_ and uses
        Jason Swail's hack from `ParmEd/ParmEd#722`_ to switch between
        :mod:`scipy.io.netcdf` and :mod:`netCDF4`.

        .. _`Issue 109`:
           https://github.com/MDAnalysis/mdanalysis/issues/109
        .. _`netcdf4storage.py`:
           https://storage.googleapis.com/google-code-attachments/mdanalysis/issue-109/comment-2/netcdf4storage.py
        .. _`ParmEd/ParmEd#722`: https://github.com/ParmEd/ParmEd/pull/722

        """
        if not self._first_frame:
            raise IOError(
                errno.EIO,
                "Attempt to write to closed file {0}".format(self.filename))

        if netCDF4:
            ncfile = netCDF4.Dataset(self.filename, 'w',
                                     format='NETCDF3_64BIT')
        else:
            ncfile = scipy.io.netcdf_file(self.filename,
                                          mode='w', version=2,
                                          maskandscale=False)
            wmsg = ("Could not find netCDF4 module. Falling back to MUCH "
                    "slower scipy.io.netcdf implementation for writing.")
            logger.warning(wmsg)
            warnings.warn(wmsg)

        # Set global attributes.
        setattr(ncfile, 'program', 'MDAnalysis.coordinates.TRJ.NCDFWriter')
        setattr(ncfile, 'programVersion', MDAnalysis.__version__)
        setattr(ncfile, 'Conventions', 'AMBER')
        setattr(ncfile, 'ConventionVersion', '1.0')
        setattr(ncfile, 'application', 'MDAnalysis')

        # Create dimensions
        ncfile.createDimension('frame',
                               None)  # unlimited number of steps (can append)
        ncfile.createDimension('atom',
                               self.n_atoms)  # number of atoms in system
        ncfile.createDimension('spatial', 3)  # number of spatial dimensions
        ncfile.createDimension('cell_spatial', 3)  # unitcell lengths
        ncfile.createDimension('cell_angular', 3)  # unitcell angles
        ncfile.createDimension('label', 5)  # needed for cell_angular

        # Create variables.
        coords = ncfile.createVariable('coordinates', 'f4',
                                       ('frame', 'atom', 'spatial'))
        setattr(coords, 'units', 'angstrom')
        if self.scale_factors['coordinates']:
            coords.scale_factor = self.scale_factors['coordinates']

        spatial = ncfile.createVariable('spatial', 'c', ('spatial', ))
        spatial[:] = np.asarray(list('xyz'))

        time = ncfile.createVariable('time', 'f4', ('frame',))
        setattr(time, 'units', 'picosecond')
        if self.scale_factors['time']:
            time.scale_factor = self.scale_factors['time']

        self.periodic = periodic
        if self.periodic:
            cell_lengths = ncfile.createVariable('cell_lengths', 'f8',
                                                 ('frame', 'cell_spatial'))
            setattr(cell_lengths, 'units', 'angstrom')
            if self.scale_factors['cell_lengths']:
                cell_lengths.scale_factor = self.scale_factors['cell_lengths']

            cell_spatial = ncfile.createVariable('cell_spatial', 'c',
                                                 ('cell_spatial', ))
            cell_spatial[:] = np.asarray(list('abc'))

            cell_angles = ncfile.createVariable('cell_angles', 'f8',
                                                ('frame', 'cell_angular'))
            setattr(cell_angles, 'units', 'degree')
            if self.scale_factors['cell_angles']:
                cell_angles.scale_factor = self.scale_factors['cell_angles']

            cell_angular = ncfile.createVariable('cell_angular', 'c',
                                                 ('cell_angular', 'label'))
            cell_angular[:] = np.asarray([list('alpha'), list('beta '), list(
                'gamma')])

        # These properties are optional, and are specified on Writer creation
        if self.has_velocities:
            velocs = ncfile.createVariable('velocities', 'f4',
                                           ('frame', 'atom', 'spatial'))
            setattr(velocs, 'units', 'angstrom/picosecond')
            if self.scale_factors['velocities']:
                velocs.scale_factor = self.scale_factors['velocities']
        if self.has_forces:
            forces = ncfile.createVariable('forces', 'f4',
                                           ('frame', 'atom', 'spatial'))
            setattr(forces, 'units', 'kilocalorie/mole/angstrom')
            if self.scale_factors['forces']:
                forces.scale_factor = self.scale_factors['forces']

        # Important for netCDF4! Disable maskandscale for created variables!
        # Won't work if called before variable creation!
        if netCDF4:
            ncfile.set_auto_maskandscale(False)

        ncfile.sync()
        self._first_frame = False
        self.trjfile = ncfile

    def is_periodic(self, ts):
        """Test if timestep ``ts`` contains a periodic box.

        Parameters
        ----------
        ts : :class:`Timestep`
             :class:`Timestep` instance containing coordinates to
             be written to trajectory file

        Returns
        -------
        bool
            Return ``True`` if `ts` contains a valid simulation box
        """
        return ts.dimensions is not None

    def _write_next_frame(self, ag):
        """Write information associated with ``ag`` at current frame into trajectory

        Parameters
        ----------
        ag : AtomGroup or Universe


        .. versionchanged:: 1.0.0
           Added ability to use either AtomGroup or Universe.
           Renamed from `write_next_timestep` to `_write_next_frame`.
        .. versionchanged:: 2.0.0
           Deprecated support for Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        try:
            # Atomgroup?
            ts = ag.ts
        except AttributeError:
            try:
                # Universe?
                ts = ag.trajectory.ts
            except AttributeError:
                errmsg = "Input obj is neither an AtomGroup or Universe"
                raise TypeError(errmsg) from None

        if ts.n_atoms != self.n_atoms:
            raise IOError(
                "NCDFWriter: Timestep does not have the correct number of atoms")

        if self.trjfile is None:
            # first time step: analyze data and open trajectory accordingly
            self._init_netcdf(periodic=self.is_periodic(ts))

        return self._write_next_timestep(ts)

    def _set_frame_var_and_scale(self, varname, data):
        """Helper function to set variables and scale them if necessary.

        Note
        ----
        If scale_factor is numerically close to 1.0, the variable data is not
        scaled.
        """
        sfactor = self.scale_factors[varname]
        if sfactor is None or isclose(sfactor, 1):
            self.trjfile.variables[varname][self.curr_frame] = data
        else:
            self.trjfile.variables[varname][self.curr_frame] = data / sfactor

    def _write_next_timestep(self, ts):
        """Write coordinates and unitcell information to NCDF file.

        Do not call this method directly; instead use
        :meth:`write` because some essential setup is done
        there before writing the first frame.

        Based on Joshua Adelman's `netcdf4storage.py`_ in `Issue 109`_.

        .. _`Issue 109`:
           https://github.com/MDAnalysis/mdanalysis/issues/109
        .. _`netcdf4storage.py`:
           https://storage.googleapis.com/google-code-attachments/mdanalysis/issue-109/comment-2/netcdf4storage.py


        .. versionchanged:: 2.0.0
           Can now write scale_factors, and scale variables accordingly.
        """
        pos = ts._pos
        time = ts.time
        unitcell = ts.dimensions

        if self.convert_units:
            # make a copy of the scaled positions so that the in-memory
            # timestep is not changed (would have lead to wrong results if
            # analysed *after* writing a time step to disk). The new
            # implementation could lead to memory problems and/or slow-down for
            # very big systems because we temporarily create a new array pos
            # for each frame written
            pos = self.convert_pos_to_native(pos, inplace=False)
            time = self.convert_time_to_native(time, inplace=False)
            unitcell = self.convert_dimensions_to_unitcell(ts)

        # write step
        # coordinates
        self._set_frame_var_and_scale('coordinates', pos)

        # time
        self._set_frame_var_and_scale('time', time)

        # unitcell
        if self.periodic:
            # cell lengths
            self._set_frame_var_and_scale('cell_lengths', unitcell[:3])

            self._set_frame_var_and_scale('cell_angles', unitcell[3:])

        # velocities
        if self.has_velocities:
            velocities = ts._velocities
            if self.convert_units:
                velocities = self.convert_velocities_to_native(
                    velocities, inplace=False)

            self._set_frame_var_and_scale('velocities', velocities)

        # forces
        if self.has_forces:
            forces = ts._forces
            if self.convert_units:
                forces = self.convert_forces_to_native(
                    forces, inplace=False)

            self._set_frame_var_and_scale('forces', forces)

        self.trjfile.sync()
        self.curr_frame += 1

    def close(self):
        if self.trjfile is not None:
            self.trjfile.close()
            self.trjfile = None


class NCDFPicklable(scipy.io.netcdf_file):
    """NetCDF file object (read-only) that can be pickled.

    This class provides a file-like object (as returned by
    :class:`scipy.io.netcdf_file`) that,
    unlike standard Python file objects,
    can be pickled. Only read mode is supported.

    When the file is pickled, filename and mmap of the open file handle in
    the file are saved. On unpickling, the file is opened by filename,
    and the mmap file is loaded.
    This means that for a successful unpickle, the original file still has to
    be accessible with its filename.


    .. note::
        This class subclasses :class:`scipy.io.netcdf_file`, please
        see the `scipy netcdf API documentation`_ for more information on
        the parameters and how the class behaviour.


    Parameters
    ----------
    filename : str or file-like
        a filename given a text or byte string.
    mmap : None or bool, optional
        Whether to mmap `filename` when reading. True when `filename`
        is a file name, False when `filename` is a file-like object.
    version : {1, 2}, optional
        Sets the netcdf file version to read / write. 1 is classic, 2 is
        64-bit offset format. Default is 1 (but note AMBER ncdf *requires* 2).
    maskandscale : bool, optional
        Whether to automatically scale and mask data. Default is False.

    Example
    -------
    ::

        f = NCDFPicklable(NCDF)
        print(f.variables['coordinates'].data)
        f.close()

    can also be used as context manager::

        with NCDFPicklable(NCDF) as f:
            print(f.variables['coordinates'].data)

    See Also
    ---------
    :class:`MDAnalysis.lib.picklable_file_io.FileIOPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.BufferIOPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.TextIOPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.GzipPicklable`
    :class:`MDAnalysis.lib.picklable_file_io.BZ2Picklable`


    .. versionadded:: 2.0.0


    .. _`scipy netcdf API documentation`: https://docs.scipy.org/doc/scipy/reference/generated/scipy.io.netcdf_file.html
    """
    def __getstate__(self):
        return (self.filename, self.use_mmap, self.version_byte,
                self.maskandscale)

    def __setstate__(self, args):
        self.__init__(args[0], mmap=args[1], version=args[2],
                      maskandscale=args[3])



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

r"""
Calculating path similarity --- :mod:`MDAnalysis.analysis.psa`
==========================================================================

:Author: Sean Seyler
:Year: 2015
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 0.10.0

.. deprecated:: 2.8.0

  This module is deprecated in favour of the mdakit
  `pathsimanalysis <https://github.com/MDAnalysis/PathSimAnalysis>`_ and
  will be removed in MDAnalysis 3.0.0.


See Also
--------
:mod:`pathsimanalysis.psa`

"""

import warnings

try:
    from pathsimanalysis import (
        get_path_metric_func,
        sqnorm,
        get_msd_matrix,
        reshaper,
        get_coord_axes,
        hausdorff,
        hausdorff_wavg,
        hausdorff_avg,
        hausdorff_neighbors,
        discrete_frechet,
        dist_mat_to_vec,
        Path,
        PSAPair,
        PSAnalysis,
    )
except ImportError:
    wmsg = (
        "Please install the PathSimAnalysis mdakit to use it in MDAnalysis.\n"
        "More details can be found here: "
        "https://www.mdanalysis.org/PathSimAnalysis/getting_started.html"
    )
    warnings.warn(wmsg, category=UserWarning)
else:
    wmsg = (
        "Deprecation in version 2.8.0:\n"
        "MDAnalysis.analysis.psa is deprecated in favour of the MDAKit "
        "PathSimAnalysis (https://github.com/MDAnalysis/PathSimAnalysis) "
        "and will be removed in MDAnalysis version 3.0.0"
    )
    warnings.warn(wmsg, category=DeprecationWarning)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import MDAnalysis as mdanalysis
from MDAnalysis.transformations import set_dimensions
from MDAnalysisTests import make_Universe


@pytest.fixture()
def boxdimensions_universe():
    # create Universe objects for tests
    new_u = make_Universe(trajectory=True)
    return new_u


@pytest.fixture()
def variable_boxdimensions_universe():
    # create Universe objects for tests
    n_atoms = 1
    n_frames = 3
    u = mdanalysis.Universe.empty(n_atoms, trajectory=True)
    coordinates = np.zeros((n_frames, u.atoms.n_atoms, 3))
    u.load_new(coordinates, order="fac")
    return u


def test_boxdimensions_dims(boxdimensions_universe):
    new_dims = np.float32([2, 2, 2, 90, 90, 90])
    set_dimensions(new_dims)(boxdimensions_universe.trajectory.ts)
    assert_array_almost_equal(
        boxdimensions_universe.dimensions, new_dims, decimal=6
    )


@pytest.mark.parametrize(
    "dim_vector_shapes",
    (
        [1, 1, 1, 90, 90],
        [1, 1, 1, 1, 90, 90, 90],
        np.array([[1], [1], [90], [90], [90]]),
        np.array([1, 1, 1, 90, 90]),
        np.array([1, 1, 1, 1, 90, 90, 90]),
        [1, 1, 1, 90, 90],
        111909090,
    ),
)
def test_dimensions_vector(boxdimensions_universe, dim_vector_shapes):
    # wrong box dimension vector shape
    ts = boxdimensions_universe.trajectory.ts
    with pytest.raises(ValueError, match="valid box dimension shape"):
        set_dimensions(dim_vector_shapes)(ts)


@pytest.mark.parametrize(
    "dim_vector_forms_dtypes",
    (
        ["a", "b", "c", "d", "e", "f"],
        np.array(["a", "b", "c", "d", "e", "f"]),
        "abcd",
    ),
)
def test_dimensions_vector_asarray(
    boxdimensions_universe, dim_vector_forms_dtypes
):
    # box dimension input type not convertible into array
    ts = boxdimensions_universe.trajectory.ts
    with pytest.raises(ValueError, match="cannot be converted"):
        set_dimensions(dim_vector_forms_dtypes)(ts)


def test_dimensions_transformations_api(boxdimensions_universe):
    # test if transformation works with on-the-fly transformations API
    new_dims = np.float32([2, 2, 2, 90, 90, 90])
    transform = set_dimensions(new_dims)
    boxdimensions_universe.trajectory.add_transformations(transform)
    for ts in boxdimensions_universe.trajectory:
        assert_array_almost_equal(
            boxdimensions_universe.dimensions, new_dims, decimal=6
        )


def test_varying_dimensions_transformations_api(
    variable_boxdimensions_universe,
):
    """
    Test if transformation works with on-the-fly transformations API
    when we have varying dimensions.
    """
    new_dims = np.float32(
        [
            [2, 2, 2, 90, 90, 90],
            [4, 4, 4, 90, 90, 90],
            [8, 8, 8, 90, 90, 90],
        ]
    )
    transform = set_dimensions(new_dims)
    variable_boxdimensions_universe.trajectory.add_transformations(transform)
    for ts in variable_boxdimensions_universe.trajectory:
        assert_array_almost_equal(
            variable_boxdimensions_universe.dimensions,
            new_dims[ts.frame],
            decimal=6,
        )


def test_varying_dimensions_no_data(
    variable_boxdimensions_universe,
):
    """
    Check error is raised when dimensions are missing for a frame
    in a trajectory.
    """
    # trjactory has three frames
    new_dims = np.float32(
        [
            [2, 2, 2, 90, 90, 90],
            [4, 4, 4, 90, 90, 90],
        ]
    )
    transform = set_dimensions(new_dims)
    with pytest.raises(
        ValueError, match="Dimensions array has no data for frame 2"
    ):
        variable_boxdimensions_universe.trajectory.add_transformations(
            transform
        )



import MDAnalysis
import numpy as np

try:
    from MDAnalysis.analysis import psa
except:
    pass


class PSA_sqnormBench(object):
    """Benchmarks for MDAnalysis.analysis.psa.
    sqnorm
    """

    params = ([2, 3, 4], [100, 1000, 10000], [None, 0, 1, -1])

    # num_cols is equivalent to dimensions
    # num_rows is equivalent to i.e., num atoms
    param_names = ["num_cols", "num_rows", "axis"]

    def setup(self, num_cols, num_rows, axis):
        np.random.seed(170089)
        self.v = np.random.rand(num_rows, num_cols)

    def time_sqnorm(self, num_cols, num_rows, axis):
        """Benchmark sqnorm in psa module"""
        psa.sqnorm(v=self.v, axis=axis)


class PSA_get_msd_matrixBench(object):
    """Benchmarks for MDAnalysis.analysis.psa.
    get_msd_matrix
    """

    params = ([10, 100, 1000], [5, 25, 50])

    # since the function is defined to work with
    # 3N dimension data sets, we will restrict
    # benchmarks to that dimensionality
    param_names = ["time_steps", "n_atoms"]

    def setup(self, time_steps, n_atoms):
        np.random.seed(170089)
        self.P = np.random.rand(time_steps, n_atoms, 3)
        np.random.seed(971132)
        self.Q = np.random.rand(time_steps, n_atoms, 3)

    def time_get_msd_matrix(self, time_steps, n_atoms):
        """Benchmark for get_msd_matrix in psa module"""
        # only default argument for axis is benchmarked
        psa.get_msd_matrix(P=self.P, Q=self.Q, axis=None)


class PSA_get_coord_axesBench(object):
    """Benchmarks for MDAnalysis.analysis.psa.
    get_coord_axes
    """

    params = ([10, 100, 1000], [5, 25, 50])

    param_names = ["time_steps", "n_atoms"]

    def setup(self, time_steps, n_atoms):
        np.random.seed(170089)
        # only using condensed path input
        # data structure for now
        self.path = np.random.rand(time_steps, n_atoms * 3)

    def time_get_coord_axes(self, time_steps, n_atoms):
        """Benchmark get_coord_axes in psa module"""
        psa.get_coord_axes(path=self.path)


class PSA_get_path_metric_funcBench(object):
    """Benchmark for MDAnalysis.analysis.psa.
    get_path_metric_func
    """

    params = [
        "hausdorff",
        "weighted_average_hausdorff",
        "average_hausdorff",
        "hausdorff_neighbors",
        "discrete_frechet",
    ]

    param_names = ["path_metric"]

    def time_get_path_metric_func(self, path_metric):
        """Benchmark for get_path_metric_func in psa
        module
        """
        psa.get_path_metric_func(name=path_metric)


class PSA_metricBench(object):
    """Benchmarks for the various path metric calculations
    in the psa module.
    """

    params = ([10, 100, 200], [5, 25, 50])

    param_names = ["time_steps", "n_atoms"]

    def setup(self, time_steps, n_atoms):
        np.random.seed(170089)
        self.P = np.random.rand(time_steps, n_atoms, 3)
        np.random.seed(971132)
        self.Q = np.random.rand(time_steps, n_atoms, 3)

    def time_hausdorff(self, time_steps, n_atoms):
        """Benchmark for hausdorff() in psa module."""
        psa.hausdorff(P=self.P, Q=self.Q)

    def time_hausdorff_wavg(self, time_steps, n_atoms):
        """Benchmark for hausdorff_wavg() in psa module."""
        psa.hausdorff_wavg(P=self.P, Q=self.Q)

    def time_hausdorff_avg(self, time_steps, n_atoms):
        """Benchmark for hausdorff_avg() in psa module."""
        psa.hausdorff_avg(P=self.P, Q=self.Q)

    def time_hausdorff_neighbors(self, time_steps, n_atoms):
        """Benchmark for hausdorff_neighbors() in psa module."""
        psa.hausdorff_neighbors(P=self.P, Q=self.Q)

    def time_discrete_frechet(self, time_steps, n_atoms):
        """Benchmark for discrete_frechet() in psa module."""
        psa.discrete_frechet(P=self.P, Q=self.Q)
